# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import base64
import logging
import struct
import time
import types
import typing
import uuid
import warnings
import xml.etree.ElementTree as ET

import requests.exceptions
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from pypsrp import _pool_manager
from pypsrp._utils import version_equal_or_newer
from pypsrp.complex_objects import (
    ApartmentState,
    Command,
    CommandMetadata,
    CommandParameter,
    CommandType,
    ComplexObject,
    HostInfo,
    ObjectMeta,
    Pipeline,
    PipelineResultTypes,
    PSInvocationState,
    PSThreadOptions,
    RemoteStreamOptions,
    RunspacePoolState,
)
from pypsrp.exceptions import (
    FragmentError,
    InvalidPipelineStateError,
    InvalidPSRPOperation,
    InvalidRunspacePoolStateError,
    WSManFaultError,
)
from pypsrp.messages import (
    ApplicationPrivateData,
    ConnectRunspacePool,
    CreatePipeline,
    DebugRecord,
    Destination,
    EncryptedSessionKey,
    EndOfPipelineInput,
    ErrorRecordMessage,
    GetAvailableRunspaces,
    GetCommandMetadata,
    InformationRecord,
    InitRunspacePool,
    Message,
    MessageType,
    PipelineHostCall,
    PipelineHostResponse,
    PipelineInput,
    PipelineState,
    ProgressRecord,
    PublicKey,
    ResetRunspaceState,
    RunspaceAvailability,
    RunspacePoolHostCall,
    RunspacePoolHostResponse,
    RunspacePoolInitData,
    RunspacePoolStateMessage,
    SessionCapability,
    SetMaxRunspaces,
    SetMinRunspaces,
    UserEvent,
    VerboseRecord,
    WarningRecord,
)
from pypsrp.serializer import Serializer
from pypsrp.shell import SignalCode, WinRS
from pypsrp.wsman import NAMESPACES, OptionSet, SelectorSet, WSMan

if typing.TYPE_CHECKING:
    from pypsrp.host import PSHost

log = logging.getLogger(__name__)

PROTOCOL_VERSION = "2.3"
PS_VERSION = "2.0"
SERIALIZATION_VERSION = "1.1.0.1"
DEFAULT_CONFIGURATION_NAME = "Microsoft.PowerShell"


class RunspacePoolWarning(Warning):
    pass


class RunspacePool(object):
    def __init__(
        self,
        connection: WSMan,
        apartment_state: int = ApartmentState.UNKNOWN,
        thread_options: int = PSThreadOptions.DEFAULT,
        host: typing.Optional["PSHost"] = None,
        configuration_name: str = DEFAULT_CONFIGURATION_NAME,
        min_runspaces: int = 1,
        max_runspaces: int = 1,
        session_key_timeout_ms: int = 60000,
        *,
        no_profile: bool = False,
        idle_timeout: int | None = None,
    ) -> None:
        """
        Represents a Runspace pool on a remote host. This pool can contain
        one or more running PowerShell instances.

        This is meant to be a near representation of the
        System.Management.Automation.Runspaces.RunspacePool .NET class

        :param connection: The connection object used to send the command over
            to the remote host. Right now only pypsrp.wsman.WSMan is supported
        :param apartment_state: The ApartmentState enum int values that specify
            the apartment state of the remote thread
        :param thread_options: The ThreadOptions enum int values that specify
            what type of thread of create
        :param host: An implementation of pypsrp.host.PSHost that defines
            the current host and what it supports
        :param configuration_name: The PSRP configuration name to connect to,
            by default this is Microsoft.PowerShell which opens a full
            PowerShell endpoint, use this to specify the JEA configuration to
            connect to
        :param min_runspaces: The minimum number of runspaces that a pool can
            hold
        :param max_runspaces: The maximum number of runspaces that a pool can
            hold
        :param session_key_timeout_ms: The maximum time to wait for a session
            key transfer from the server
        :param no_profile: If True, the user profile will not be loaded and
            will use the machine defaults.
        :param idle_timeout: The idle timeout, in seconds, for the runspace. The
            runspace will be closed if no operations are performed in the time
            specified. If None the server default will be used.
        """
        log.info("Initialising RunspacePool object for configuration %s" % configuration_name)
        # The below are defined in some way at
        # https://msdn.microsoft.com/en-us/library/ee176015.aspx
        self.id = str(uuid.uuid4()).upper()
        self.state = RunspacePoolState.BEFORE_OPEN
        self.connection = connection
        resource_uri = "http://schemas.microsoft.com/powershell/%s" % configuration_name
        self.shell = WinRS(
            connection,
            resource_uri=resource_uri,
            id=self.id,
            input_streams="stdin pr",
            output_streams="stdout",
            no_profile=no_profile,
            idle_time_out=idle_timeout,
        )
        self.ci_table: typing.Dict = {}
        self.pipelines: typing.Dict[str, "PowerShell"] = {}
        self.session_key_timeout_ms = session_key_timeout_ms

        # Extra properties that are important and can control the RunspacePool
        # behaviour
        self.apartment_state = apartment_state
        self.thread_options = thread_options
        self.host = host
        self.protocol_version: typing.Optional[str] = None
        self.ps_version: typing.Optional[str] = None
        self.serialization_version: typing.Optional[str] = None
        self.user_events: typing.List[UserEvent] = []

        self._application_private_data: typing.Optional[ApplicationPrivateData] = None
        self._min_runspaces = min_runspaces
        self._max_runspaces = max_runspaces
        self._serializer = Serializer()
        self._fragmenter = Fragmenter(self.connection.max_payload_size, self._serializer)
        self._exchange_key: typing.Optional[rsa.RSAPrivateKeyWithSerialization] = None
        self._key_exchanged = False
        self._new_client = False
        self._ci_counter = 0

    @property
    def application_private_data(self) -> typing.Optional[ApplicationPrivateData]:
        """
        Private data to be used by applications built on top of PowerShell.
        Runspace data is gathered when creating the remote runspace pool and
        will be None if the runspace is not connected.
        """
        return self._application_private_data

    @property
    def min_runspaces(self) -> int:
        return self._min_runspaces

    @min_runspaces.setter
    def min_runspaces(
        self,
        min_runspaces: int,
    ) -> None:
        """
        Sets the minimum number of Runspaces that the pool maintains in
        anticipation of new requests.

        :param min_runspaces: The minimum number of runspaces in the pool
        """
        log.info("Setting minimum runspaces for pool to %d" % min_runspaces)
        if self.state != RunspacePoolState.OPENED:
            self._min_runspaces = min_runspaces
            return
        elif min_runspaces == self._min_runspaces:
            return

        def response_handler(response):
            if not response:
                raise InvalidPSRPOperation("Failed to set minimum runspaces")
            return response

        ci = self._ci_counter
        self._ci_counter += 1
        self.ci_table[ci] = response_handler

        set_min_runspace = SetMinRunspaces(min_runspaces=min_runspaces, ci=ci)
        data = self._fragmenter.fragment(set_min_runspace, self.id)[0]
        self.shell.send("stdin", data)

        while not isinstance(self.ci_table[ci], bool):
            self._receive()
        self._min_runspaces = min_runspaces
        del self.ci_table[ci]

    @property
    def max_runspaces(self) -> int:
        return self._max_runspaces

    @max_runspaces.setter
    def max_runspaces(
        self,
        max_runspaces: int,
    ) -> None:
        """
        Sets the maximum number of Runspaces that can be active concurrently
        in the pool. All requests above that number remain queued until
        runspaces become available.

        :param max_runspaces: The maximum number of runspaces in the pool
        """
        log.info("Setting maximum runspaces for pool to %d" % max_runspaces)
        if self.state != RunspacePoolState.OPENED:
            self._max_runspaces = max_runspaces
            return
        elif max_runspaces == self._max_runspaces:
            return

        def response_handler(response):
            if not response:
                raise InvalidPSRPOperation("Failed to set maximum runspaces")
            return response

        ci = self._ci_counter
        self._ci_counter += 1
        self.ci_table[ci] = response_handler

        set_max_runspace = SetMaxRunspaces(max_runspaces=max_runspaces, ci=ci)
        data = self._fragmenter.fragment(set_max_runspace, self.id)[0]
        self.shell.send("stdin", data)

        while not isinstance(self.ci_table[ci], bool):
            self._receive()
        self._max_runspaces = max_runspaces
        del self.ci_table[ci]

    def __enter__(self) -> "RunspacePool":
        self.open()
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> None:
        self.close()

    def close(self) -> None:
        """
        Closes the RunspacePool and cleans all the internal resources. This
        will close all the runspaces in the runspacepool and release all the
        operations waiting for a runspace. If the pool is already closed or
        broken or closing, this will just return
        """
        log.info("Closing Runspace Pool")
        if self.state in [RunspacePoolState.CLOSED, RunspacePoolState.CLOSING, RunspacePoolState.BROKEN]:
            return

        self.shell.close()
        self.state = RunspacePoolState.CLOSED

    def connect(self) -> None:
        """
        Connects the runspace pool, Runspace pool must be in a disconnected
        state. This only supports reconnecting to a runspace pool created by
        the same client with the same SessionId value in the WSMan headers.
        """
        log.info("Connecting to Runspace Pool")
        if self.state == RunspacePoolState.OPENED:
            return
        elif self.state != RunspacePoolState.DISCONNECTED:
            raise InvalidRunspacePoolStateError(
                self.state, RunspacePoolState.DISCONNECTED, "connect to a disconnected Runspace Pool"
            )

        if self._new_client:
            log.debug("Remote Runspace Pool was created with a different client, connecting as new client")
            self._connect_new_client()
            self._new_client = False
        else:
            log.debug("Remote Runspace Pool was created with the same client, connecting as an existing client")
            self._connect_existing_client()

    def _connect_existing_client(self) -> None:
        selector_set = SelectorSet()
        selector_set.add_option("ShellId", self.shell.id or "")

        self.connection.reconnect(self.shell.resource_uri, selector_set=selector_set)
        self.state = RunspacePoolState.OPENED

    def _connect_new_client(self) -> None:
        rsp = NAMESPACES["rsp"]
        session_capability = SessionCapability(PROTOCOL_VERSION, PS_VERSION, SERIALIZATION_VERSION)
        connect_runspace = ConnectRunspacePool()
        data = self._fragmenter.fragment(session_capability, self.id)[0]
        data += self._fragmenter.fragment(connect_runspace, self.id)[0]

        connect = ET.Element("{%s}Connect" % rsp)
        selectors = SelectorSet()
        selectors.add_option("ShellId", self.id)

        options = OptionSet()
        options.add_option("protocolversion", PROTOCOL_VERSION, {"MustComply": "true"})

        open_content = ET.SubElement(connect, "connectXml", xmlns="http://schemas.microsoft.com/powershell")
        open_content.text = base64.b64encode(data).decode("utf-8")

        response = self.connection.connect(self.shell.resource_uri, connect, options, selectors)
        response_xml = response.find("rsp:ConnectResponse/pwsh:connectResponseXml", NAMESPACES)
        fragments = base64.b64decode(response_xml.text)  # type: ignore[arg-type, union-attr] # Mandated by spec

        self._parse_responses(fragments)
        self.shell.id = self.id  # need to sync up the ShellID with the rs ID
        self.shell._selector_set = selectors
        self._receive()

    def create_disconnected_power_shells(self) -> typing.List["PowerShell"]:
        """
        Creates a list of PowerShell objects that are in the Disconnected state
        for all currently disconnected running commands associated with this
        runspace pool.

        :return: List<PowerShell>: List of disconnected PowerShell objects
        """
        log.info("Getting list of disconnected PowerShells for the current Runspace Pool")
        return [s for s in self.pipelines.values() if s.state == PSInvocationState.DISCONNECTED]

    def disconnect(self) -> None:
        """
        Disconnects the runspace pool, must be in the Opened state
        """
        log.info("Disconnecting from Runspace Pool")
        if self.state == RunspacePoolState.DISCONNECTED:
            return
        elif self.state != RunspacePoolState.OPENED:
            raise InvalidRunspacePoolStateError(self.state, RunspacePoolState.OPENED, "disconnect a Runspace Pool")

        disconnect = ET.Element("{%s}Disconnect" % NAMESPACES["rsp"])
        selector_set = SelectorSet()
        selector_set.add_option("ShellId", self.shell.id or "")

        self.connection.disconnect(self.shell.resource_uri, disconnect, selector_set=selector_set)
        self.state = RunspacePoolState.DISCONNECTED
        for pipeline in self.pipelines.values():
            pipeline.state = PSInvocationState.DISCONNECTED

    def get_available_runspaces(self) -> int:
        """
        Retrieves the number of runspaces available at the time of calling this
        method.

        :return: The number of available runspaces in the pool
        """

        def response_handler(response):
            self._max_runspaces = response
            return response

        ci = self._ci_counter
        self._ci_counter += 1
        self.ci_table[ci] = response_handler

        get_runspaces = GetAvailableRunspaces(ci=ci)
        data = self._fragmenter.fragment(get_runspaces, self.id)[0]
        self.shell.send("stdin", data)

        avail_runspaces = None
        while avail_runspaces is None:
            self._receive()
            if isinstance(self.ci_table[ci], int):
                avail_runspaces = self.ci_table[ci]

        del self.ci_table[ci]
        return avail_runspaces

    def get_command_metadata(
        self,
        names: typing.Union[str, typing.List[str]],
        command_types: int = CommandType.ALL,
        namespace: typing.Optional[typing.List[str]] = None,
        arguments: typing.Optional[typing.List] = None,
    ) -> typing.List[CommandMetadata]:
        """
        Get's metadata of the higher layer command's. This is very similar to
        running the Get-Help cmdlet's but just done on the protocol layer
        instead of through a manual PowerShell pipeline.

        Any wildcard of escaping of strings must conform to the MS-PSRP
        standard which is defined here
        https://msdn.microsoft.com/en-us/library/ee175957.aspx

        :param names: A string or list of strings that specify the commands to
            get the metadata for. Each string can be a wildcard instead of the
            full name.
        :param command_types: The complex_objects.CommandType to filter by, by
            default this will be CommandType.ALL
        :param namespace: A list of wildcard patterns describing the command
            namespaces containing the commands that the server should return
        :param arguments: A list of extra arguments passed to the higher-layer
            above the PowerShell Remoting Protocol
        :return: List of CommandMetadata objects returned by the server
        """
        if self.state != RunspacePoolState.OPENED:
            raise InvalidRunspacePoolStateError(self.state, RunspacePoolState.OPENED, "get command metadata")

        if not isinstance(names, list):
            names = [names]

        get_msg = GetCommandMetadata(names, command_types, namespace, arguments)
        ps = PowerShell(self)
        ps._invoke(get_msg)
        output = ps.end_invoke()

        # first output obj is the MetadataCount obj that tells us how many
        # CommandMetadata objects there are
        _ = output.pop(0)

        # because the main type is generic the serializer will output a
        # GenericComplexObject, we will just manually serialize them to the
        # CommandMetadata object
        meta = ObjectMeta("Obj", object=CommandMetadata)
        return [self._serializer.deserialize(o._xml, meta) for o in output]

    @staticmethod
    def get_runspace_pools(
        connection: WSMan,
    ) -> typing.List["RunspacePool"]:
        """
        Queries the server for disconnected runspace pools and creates a list
        of runspace pool objects associated with each disconnected runspace
        pool on the server. Each runspace pool object in the returned array is
        in the Disconnected state and can be connected to the server by calling
        the connect() method on the runspace pool.

        :param connection: The connection object used to query the remote
            server for a list of Runspace Pools. Currently only
            pypsrp.wsman.WSMan is supported.
        :return: List<RunspacePool> objects each in the Disconnected state
        """
        log.info("Getting list of Runspace Pools on remote host")
        wsen = NAMESPACES["wsen"]
        wsmn = NAMESPACES["wsman"]

        enum_msg = ET.Element("{%s}Enumerate" % wsen)
        ET.SubElement(enum_msg, "{%s}OptimizeEnumeration" % wsmn)
        ET.SubElement(enum_msg, "{%s}MaxElements" % wsmn).text = "32000"

        # TODO: support wsman:EndOfSequence
        response = connection.enumerate("http://schemas.microsoft.com/wbem/wsman/1/windows/shell", enum_msg)
        shells = response.findall("wsen:EnumerateResponse/wsman:Items/rsp:Shell", NAMESPACES)

        runspace_pools = []
        for shell in shells:
            shell_id: str = shell.find("rsp:ShellId", NAMESPACES).text  # type: ignore # Mandated by spec
            pool = RunspacePool(connection)
            pool.id = shell_id
            pool.shell.id = shell_id
            pool.shell.opened = True
            pool._new_client = True

            # Seems like the server sends all pools not just disconnected but
            # the .NET API always sets the state to Disconnected when callling
            # GetRunspacePools so we replicate that here
            pool.state = RunspacePoolState.DISCONNECTED

            enum_msg = ET.Element("{%s}Enumerate" % wsen)
            ET.SubElement(enum_msg, "{%s}OptimizeEnumeration" % wsmn)
            ET.SubElement(enum_msg, "{%s}MaxElements" % wsmn).text = "32000"
            filter = ET.SubElement(
                enum_msg, "{%s}Filter" % wsmn, Dialect="http://schemas.dmtf.org/wbem/wsman/1/wsman/SelectorFilter"
            )
            selector_set = SelectorSet()
            selector_set.add_option("ShellId", shell_id)
            filter.append(selector_set.pack())

            resp = connection.enumerate("http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Command", enum_msg)
            commands = resp.findall("wsen:EnumerateResponse/wsman:Items/rsp:Command", NAMESPACES)
            pipelines = {}
            for command in commands:
                command_id: str = command.find("rsp:CommandId", NAMESPACES).text  # type: ignore # Mandated by spec

                powershell = PowerShell(pool)
                powershell.id = command_id
                powershell._command_id = command_id
                powershell.state = PSInvocationState.DISCONNECTED
                pipelines[powershell.id] = powershell

            pool.pipelines = pipelines
            runspace_pools.append(pool)
        return runspace_pools

    def open(
        self,
        application_arguments: typing.Optional[typing.Dict] = None,
    ) -> None:
        """
        Opens the runspace pool, this step must be called before it can be
        used.

        :param application_arguments: A dictionary of variables to set for the
            runspace host. These can then be accessed in a PowerShell instance
            of the runspace with $PSSenderInfo.ApplicationArguments
        """
        log.info("Opening a new Runspace Pool on remote host")
        if self.state == RunspacePoolState.OPENED:
            return
        if self.state != RunspacePoolState.BEFORE_OPEN:
            raise InvalidRunspacePoolStateError(self.state, RunspacePoolState.BEFORE_OPEN, "open a new Runspace Pool")

        session_capability = SessionCapability(PROTOCOL_VERSION, PS_VERSION, SERIALIZATION_VERSION)
        init_runspace_pool = InitRunspacePool(
            self.min_runspaces,
            self.max_runspaces,
            PSThreadOptions(value=self.thread_options),
            ApartmentState(value=self.apartment_state),
            HostInfo(host=self.host),
            application_arguments,
        )
        msgs = [session_capability, init_runspace_pool]
        fragments = self._fragmenter.fragment_multiple(msgs, self.id)

        open_content = ET.Element("creationXml", xmlns="http://schemas.microsoft.com/powershell")
        open_content.text = base64.b64encode(fragments.pop(0)).decode("utf-8")

        options = OptionSet()
        options.add_option("protocolversion", PROTOCOL_VERSION, {"MustComply": "true"})
        self.shell.open(options, open_content)
        self.state = RunspacePoolState.NEGOTIATION_SENT

        responses = []
        while self.state == RunspacePoolState.NEGOTIATION_SENT:
            responses.extend(self._receive())

    def exchange_keys(self) -> None:
        """
        Initiate a key exchange with the server that is required when dealing
        with secure strings. This can only be run once the RunspacePool is
        open and if the key has already been exchanged then nothing will
        happen.
        """
        log.info("Starting key exchange with remote host")
        if self._key_exchanged or self._exchange_key is not None:
            # key is already exchanged or we are still in the processes of
            # exchanging it, no need to run again
            return

        # Generate a unique RSA key pair for use in this Pool only
        self._exchange_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )
        public_numbers = self._exchange_key.public_key().public_numbers()
        exponent = struct.pack("<I", public_numbers.e)
        modulus = b""
        for i in range(0, 256):
            byte_value = struct.pack("B", public_numbers.n >> (i * 8) & 0xFF)
            modulus += byte_value

        # the public key bytes follow a set structure defined in MS-PSRP
        public_key_bytes = b"\x06\x02\x00\x00\x00\xa4\x00\x00\x52\x53\x41\x31\x00\x08\x00\x00" + exponent + modulus
        public_key = base64.b64encode(public_key_bytes)

        msg = PublicKey(public_key=public_key.decode("utf-8"))
        fragments = self._fragmenter.fragment(msg, self.id)
        for fragment in fragments:
            self.shell.send("stdin", fragment)

        start = time.time()
        while not self._key_exchanged:
            elapsed = int((time.time() - start) * 1000)
            if elapsed > self.session_key_timeout_ms:
                raise InvalidPSRPOperation("Timeout while waiting for key exchange")
            self._receive()

    def reset_runspace_state(self) -> None:
        """
        Resets the variable table for the runspace to the default state.

        This is only supported for the protocol version 2.3 and above (Server
        2016/Windows 10+)
        """
        log.info("Resetting remote Runspace Pool state")
        if self.state == RunspacePoolState.BEFORE_OPEN:
            # no need to reset if the runspace has not been opened
            return
        elif self.state != RunspacePoolState.OPENED:
            raise InvalidRunspacePoolStateError(
                self.state, [RunspacePoolState.BEFORE_OPEN, RunspacePoolState.OPENED], "reset RunspacePool state"
            )
        elif not version_equal_or_newer(self.protocol_version or "", "2.3"):
            raise InvalidPSRPOperation(
                "Cannot reset runspace state on "
                "protocol versions older than 2.3, "
                "actual: %s" % self.protocol_version
            )

        def response_handler(response):
            if not response:
                raise InvalidPSRPOperation("Failed to reset runspace state")
            return response

        ci = self._ci_counter
        self._ci_counter += 1
        self.ci_table[ci] = response_handler

        reset_state = ResetRunspaceState(ci=ci)
        data = self._fragmenter.fragment(reset_state, self.id)[0]
        self.shell.send("stdin", data)

        while not isinstance(self.ci_table[ci], bool):
            self._receive()
        del self.ci_table[ci]

    def serialize(
        self,
        obj: typing.Any,
        metadata: typing.Optional[ObjectMeta] = None,
    ) -> typing.Optional[ET.Element]:
        """
        Serialize a Python object to PSRP object. This can try to automatically
        serialize based on the Python type to the closest PSRP object but
        manual coercion can be done with the metadata parameter.

        :param obj: The Python object to serialize
        :param metadata: complex_objects.ObjectMeta that defines the type of
            object to serialize to, if omitted the obj will be serialized based
            on the Python type
        :return: An XML element that can be used as part of the PSRP input
            elements like cmdlet parameters
        """
        return self._serializer.serialize(obj, metadata=metadata)

    def is_alive(
        self,
        *,
        timeout: int | None = None,
    ) -> bool:
        """Checks whether the RunspacePool is still alive.

        :param timeout: Override the connection timeout defaults for the
            request. The WSMan operation timeout will be set to this value
            and the HTTP connect/read timeouts will be this value + 2 seconds.
        :return: A bool True if the pool is still alive, False otherwise.
        """
        is_closed = False

        try:
            # We don't want to try and open a new connection if the socket
            # was closed, we treat it as the pool is not alive.
            with _pool_manager.DisableNewConnectionsContext():
                state = self.shell._get_shell(timeout=timeout)

            is_closed = state.get("State", "") == "Disconnected"
        except WSManFaultError as exc:
            if exc.code in [
                0x80338029,  # ERROR_WSMAN_OPERATION_TIMEDOUT
                0x8033805B,  # ERROR_WSMAN_UNEXPECTED_SELECTORS
            ]:
                is_closed = True
            else:
                raise

        except (_pool_manager.NewConnectionDisabled, TimeoutError, requests.exceptions.ConnectionError):
            # If a timeout or connection error occurs, treat the pool as closed.
            is_closed = True

        if is_closed:
            # Ensures that close() doesn't try and close an already closed pool
            self.state = RunspacePoolState.CLOSED
            return False

        return True

    def _receive(
        self,
        id: typing.Optional[str] = None,
        timeout: typing.Optional[int] = None,
    ) -> typing.List[typing.Tuple[int, typing.Any]]:
        """
        Sends a Receive WSMV request to the host and processes the messages
        that are received from the host (if there are any).

        :param id: If the receive is targeted to a Pipeline then this should be
            the ID of that pipeline, if None then the receive is targeted to
            the RunspacePool
        :param timeout: An override that specifies the operation timeout for
            the receive command
        :return: List of tuples where each tuple is a tuple of
            MessageType: The Message ID of the response
            Response: The return object of the response handler function for
                the message type
        """
        command_id = None
        pipeline = None

        if id:
            pipeline = self.pipelines.get(id)
            if pipeline is not None:
                command_id = pipeline._command_id

        response = self.shell.receive("stdout", command_id=command_id, timeout=timeout)[2]["stdout"]
        return self._parse_responses(response, pipeline)

    def _parse_responses(
        self,
        responses: bytes,
        pipeline: typing.Optional["PowerShell"] = None,
    ) -> typing.List[typing.Tuple[int, typing.Any]]:
        messages = self._fragmenter.defragment(responses)

        response_functions: typing.Dict[int, typing.Optional[typing.Callable[[Message], typing.Any]]] = {
            # While the docs say we should verify, they are out of date with
            # the possible responses and so we will just ignore for now
            MessageType.SESSION_CAPABILITY: self._process_session_capability,
            MessageType.ENCRYPTED_SESSION_KEY: self._process_encrypted_session_key,
            MessageType.PUBLIC_KEY_REQUEST: lambda m: self.exchange_keys,
            MessageType.RUNSPACEPOOL_INIT_DATA: self._process_runspacepool_init_data,
            MessageType.RUNSPACE_AVAILABILITY: self._process_runspacepool_availability,
            MessageType.RUNSPACEPOOL_STATE: self._process_runspacepool_state,
            MessageType.USER_EVENT: self._process_user_event,
            MessageType.APPLICATION_PRIVATE_DATA: self._process_application_private_data,
            MessageType.RUNSPACEPOOL_HOST_CALL: self._process_runspacepool_host_call,
            MessageType.WARNING_RECORD: self._process_runspacepool_warning,
        }

        if pipeline is not None:
            pipeline_response_functions: typing.Dict[int, typing.Optional[typing.Callable[[Message], typing.Any]]] = {
                # The Pipeline Output isn't processes and just returned back to
                # the receive caller
                MessageType.PIPELINE_OUTPUT: None,
                MessageType.ERROR_RECORD: pipeline._process_error_record,
                MessageType.PIPELINE_STATE: pipeline._process_pipeline_state,
                MessageType.DEBUG_RECORD: pipeline._process_debug_record,
                MessageType.VERBOSE_RECORD: pipeline._process_verbose_record,
                MessageType.WARNING_RECORD: pipeline._process_warning_record,
                MessageType.PROGRESS_RECORD: pipeline._process_progress_record,
                MessageType.INFORMATION_RECORD: pipeline._process_information_record,
                MessageType.PIPELINE_HOST_CALL: pipeline._process_pipeline_host_call,
            }
            response_functions.update(pipeline_response_functions)

        return_values = []
        for message in messages:
            if message.message_type not in response_functions:
                log.warning("Unsupported message type '%s' received" % message.message_type)
                response_function = None

            else:
                response_function = response_functions[message.message_type]

            if response_function is not None:
                return_value = response_function(message)
                return_values.append((message.message_type, return_value))
            else:
                return_values.append((message.message_type, message))

        return return_values

    def _process_host_call(
        self,
        message: Message,
        response_obj: typing.Union[typing.Type[RunspacePoolHostResponse], typing.Type[PipelineHostResponse]],
        pipeline: typing.Optional["PowerShell"] = None,
    ) -> ComplexObject:
        host_call = typing.cast(typing.Union[RunspacePoolHostCall, PipelineHostCall], message.data)

        if self.host is None:
            log.warning(
                "Cannot run host call as no host was defined for the "
                "runspace, method: %s, args: %s" % (str(host_call.mi), str(host_call.mp))
            )
            return host_call

        response = self.host.run_method(host_call.mi, host_call.mp, self, pipeline)

        if response is not None:
            msg = response_obj()
            msg.ci = host_call.ci
            msg.mi = host_call.mi
            msg.mr = response

            pid = pipeline.id if pipeline is not None else None
            fragments = self._fragmenter.fragment(msg, self.id, pid)
            for fragment in fragments:
                self.shell.send("pr", fragment)

        return host_call

    def _process_session_capability(
        self,
        message: Message,
    ) -> None:
        msg = typing.cast(SessionCapability, message.data)

        log.debug(
            "Received SessionCapability with protocol version: "
            "%s, ps version: %s, serialization version: %s"
            % (msg.protocol_version, msg.ps_version, msg.serialization_version)
        )
        self.protocol_version = msg.protocol_version
        self.ps_version = msg.ps_version
        self.serialization_version = msg.serialization_version

        # if protocol_version >= 2.2 and the max_envelope_size is still the
        # default, update the envelope size to 500KiB
        if version_equal_or_newer(msg.protocol_version or "", "2.2") and self.connection.max_envelope_size == 153600:
            self.connection.update_max_payload_size(512000)
            self._fragmenter.max_size = self.connection.max_payload_size

    def _process_runspacepool_init_data(
        self,
        message: Message,
    ) -> None:
        msg = typing.cast(RunspacePoolInitData, message.data)

        log.debug(
            "Received RunspacePoolInitData with min runspaces: %d and "
            "max runspaces: %d" % (msg.min_runspaces, msg.max_runspaces)
        )
        self._min_runspaces = msg.min_runspaces
        self._max_runspaces = msg.max_runspaces

    def _process_runspacepool_availability(
        self,
        message: Message,
    ) -> ComplexObject:
        msg = typing.cast(RunspaceAvailability, message.data)

        ci = msg.ci
        response = msg.response
        ci_handler = self.ci_table[ci]
        response = ci_handler(response)
        self.ci_table[ci] = response
        return response

    def _process_runspacepool_host_call(
        self,
        message: Message,
    ) -> ComplexObject:
        return self._process_host_call(message, RunspacePoolHostResponse)

    def _process_runspacepool_state(
        self,
        message: Message,
    ) -> ComplexObject:
        state_msg = typing.cast(RunspacePoolStateMessage, message.data)

        log.debug("Received RunspacePoolState with state: %d" % state_msg.state)
        self.state = state_msg.state
        if self.state == RunspacePoolState.BROKEN:
            raise InvalidPSRPOperation("Received a broken RunspacePoolState message: %s" % str(state_msg.error_record))
        return state_msg

    def _process_runspacepool_warning(
        self,
        message: Message,
    ) -> None:
        warnings.warn(str(message.data), RunspacePoolWarning)

    def _process_application_private_data(
        self,
        message: Message,
    ) -> None:
        self._application_private_data = typing.cast(ApplicationPrivateData, message.data)

    def _process_encrypted_session_key(
        self,
        message: Message,
    ) -> None:
        key_msg = typing.cast(EncryptedSessionKey, message.data)

        log.debug("Received EncryptedSessionKey response")
        enc_sess_key = base64.b64decode(key_msg.session_key)

        # strip off Win32 Crypto Blob Header and reverse the bytes
        encrypted_key = enc_sess_key[12:][::-1]
        pad_method = padding.PKCS1v15()
        decrypted_key = self._exchange_key.decrypt(encrypted_key, pad_method)  # type: ignore[union-attr] # Will be set

        iv = b"\x00" * 16  # PSRP doesn't use an IV
        algorithm = algorithms.AES(decrypted_key)
        mode = modes.CBC(iv)
        cipher = Cipher(algorithm, mode)

        self._serializer.cipher = cipher
        self._key_exchanged = True
        self._exchange_key = None

    def _process_user_event(
        self,
        message: Message,
    ) -> None:
        user_event = typing.cast(UserEvent, message.data)
        self.user_events.append(user_event)


class PowerShell(object):
    def __init__(
        self,
        runspace_pool: RunspacePool,
    ) -> None:
        """
        Represents a PowerShell command or script to execute against a
        RunspacePool.

        This is meant to be a near representation of the
        System.Management.Automation.PowerShell .NET class

        :param runspace_pool: The RunspacePool that the PowerShell instance
            will run over
        """
        log.info("Initialising PowerShell in remote Runspace Pool")
        self.runspace_pool = runspace_pool
        self.state = PSInvocationState.NOT_STARTED

        self.commands: typing.List[Command] = []
        self.had_errors = False
        self.history_string = None
        self.id = str(uuid.uuid4()).upper()
        self.is_nested = False
        self.streams = PSDataStreams()
        self.output: typing.List = []
        self._from_disconnect = False

        # this is not necessarily the same ID as id, this relates to the WSMan
        # CommandID that is created and we need to reference in the WSMan msgs
        self._command_id: typing.Optional[str] = None

    def __enter__(self) -> "PowerShell":
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        value: typing.Optional[BaseException],
        traceback: typing.Optional[types.TracebackType],
    ) -> None:
        self.close()

    def add_argument(
        self,
        value: typing.Any,
    ) -> "PowerShell":
        """
        Adds an argument to the last added command.

        :param value: The argument to add. If the value is a native Python
            type then it will be automatically serialized, otherwise if it is
            an already serialized object then that value will be used instead
        :return: The current PowerShell object with the argument added to the
            last added Command
        """
        command_parameter = CommandParameter(value=value)
        self.commands[-1].args.append(command_parameter)
        return self

    def add_command(
        self,
        command: Command,
    ) -> "PowerShell":
        """
        Add a Command object to the current command pipeline.

        :param command: Command to add
        :return: The current PowerShell object with the Command added
        """
        self.commands.append(command)
        return self

    def add_cmdlet(
        self,
        cmdlet: str,
        use_local_scope: typing.Optional[bool] = None,
    ) -> "PowerShell":
        """
        Add a cmdlet/command to the current command pipeline. This is similar
        to add_command but it takes in a string and constructs the Command
        object for you. For example to construct "Get-Process | Sort-Object"

        .add_cmdlet("Get-Process").add_cmdlet("Sort-Object")

        :param cmdlet: A string representing the cmdlet to add
        :param use_local_scope: Run the cmdlet under the local scope
        :return: The current PowerShell object with the cmdlet added
        """
        command = Command(
            cmdlet,
            protocol_version=self.runspace_pool.protocol_version or "",
            is_script=False,
            use_local_scope=use_local_scope,
        )
        self.commands.append(command)
        return self

    def add_parameter(
        self,
        parameter_name: str,
        value: typing.Any = None,
    ) -> "PowerShell":
        """
        Add a parameter to the last added command. For example to construct a
        command string "get-service -name service-name"

        .add_command("get-service").add_parameter("name", "service-name")

        :param parameter_name: The name of the parameter
        :param value: The value for the parameter, None means no value is set.
            If the value is a native Python type then it will be automatically
            serialized, otherwise if it is an already serialized object then
            that value will be used instead
        :return: the current PowerShell instance with the parameter added
        """
        command_parameter = CommandParameter(name=parameter_name, value=value)
        self.commands[-1].args.append(command_parameter)
        return self

    def add_parameters(
        self,
        parameters: typing.Dict[str, typing.Any],
    ) -> "PowerShell":
        """
        Adds a set of parameters to the last added command.

        :param parameters: A dictionary of parameters where the key is the
            parameter name and the value is the parameter value. A value of
            None means no value is set and the parameter is a switch
        :return: the current PowerShell instance with the parameters added
        """
        for parameter_name, value in parameters.items():
            self.add_parameter(parameter_name, value)
        return self

    def add_script(
        self,
        script: str,
        use_local_scope: typing.Optional[bool] = None,
    ) -> "PowerShell":
        """
        Add a piece of script to construct a command pipeline.

        :param script: A string representing a script
        :param use_local_scope: Run the script under the local scope
        :return: the current PowerShell instance with the command added
        """
        command = Command(
            script,
            protocol_version=self.runspace_pool.protocol_version or "",
            is_script=True,
            use_local_scope=use_local_scope,
        )
        self.commands.append(command)
        return self

    def add_statement(self) -> "PowerShell":
        """
        Set's the last command in the pipeline to be the last in that
        statement/pipeline so the next command is in a new statement.

        :return: The current PowerShell instance with the last command set
            as the last one in that statement
        """
        self.commands[-1].end_of_statement = True
        return self

    def clear_commands(self) -> "PowerShell":
        """
        Clears all the commands of the current PowerShell object.
        :return: The current PowerShell instance with all the commands cleared
        """
        self.commands = []
        return self

    def clear_streams(self) -> None:
        """
        Clears all the data streams of the current PowerShell object.
        """
        self.streams = PSDataStreams()
        self.output = []

    def close(self) -> None:
        """
        Closes the PowerShell pipeline and cleans up server resources.

        This method must be called to reuse a PowerShell object for multiple
        invocations. It sends a TERMINATE signal to the server and removes
        the pipeline from the runspace pool's pipeline registry.

        :raises InvalidPipelineStateError: If the pipeline is not in a terminal
            state (COMPLETED, STOPPED, or FAILED)
        """
        valid_states = [PSInvocationState.COMPLETED, PSInvocationState.STOPPED, PSInvocationState.FAILED]
        if self.state == PSInvocationState.NOT_STARTED:
            return
        elif self.state not in valid_states:
            raise InvalidPipelineStateError(self.state, valid_states, "close a PowerShell pipeline")

        self.runspace_pool.shell.signal(SignalCode.TERMINATE, command_id=self._command_id or self.id)
        self.runspace_pool.pipelines.pop(self.id, None)

    def connect(self):
        """
        Connects to a running command on a remote server, waits until the
        command is finished and returns the output objects.

        :return: Command output as a PSDataCollection
        """
        self.connect_async()
        return self.end_invoke()

    def connect_async(self):
        """
        Connects to a running command on a remote server, this method will
        connect to the host but will not wait until the command is finished.
        Call end_invoke() to wait until the process is complete.
        """
        if self.state != PSInvocationState.DISCONNECTED:
            raise InvalidPipelineStateError(
                self.state, PSInvocationState.DISCONNECTED, "connect to a disconnected pipeline"
            )
        rsp = NAMESPACES["rsp"]

        connect = ET.Element("{%s}Connect" % rsp, CommandId=self.id)

        self.runspace_pool.connection.connect(
            self.runspace_pool.shell.resource_uri, connect, selector_set=self.runspace_pool.shell._selector_set
        )
        self.state = PSInvocationState.RUNNING
        self._from_disconnect = True

    def create_nested_power_shell(self):
        """
        Creates a nested PowerShell within the current instance. Nested
        PowerShell is used to do simple operations like checking state of a
        variable while another command is using the runspace.

        Nested PowerShell should be invoked from the same thread as the parent
        PowerShell invocation thread. So effectively the parent PowerShell
        invocation thread is blocked until the nested invoke() operation is
        complete.

        :return: The new nested PowerShell object
        """
        if self.state != PSInvocationState.RUNNING:
            raise InvalidPipelineStateError(
                self.state, PSInvocationState.RUNNING, "create a nested PowerShell pipeline"
            )
        elif self._from_disconnect:
            raise InvalidPSRPOperation(
                "Cannot created a nested PowerShell "
                "pipeline from an existing pipeline "
                "that was connected to remotely"
            )

        ps = PowerShell(self.runspace_pool)
        ps.is_nested = True
        return ps

    def begin_invoke(
        self,
        input: typing.Optional[typing.Union[bytes, typing.Iterator]] = None,
        add_to_history: bool = False,
        apartment_state: typing.Optional[int] = None,
        redirect_shell_error_to_out: bool = False,
        remote_stream_options: typing.Optional[int] = RemoteStreamOptions.ADD_INVOCATION_INFO,
    ) -> None:
        """
        Invoke the command asynchronously, use end_invoke to get the output
        collection of return objects.

        :param input: List of inputs to the command, this will be manually
            serialized but can also be a result of runspace_pool.serialize()
        :param add_to_history: Add the commands run to the pool history
        :param apartment_state: Override the RunspacePool apartment state with
            one just for this invocation
        :param redirect_shell_error_to_out: Whether to redirect the global
            error output pipe to the commands error output pipe.
        :param remote_stream_options: Whether to return the invocation info on
            the various steams, see complex_objects.RemoteStreamOptions for the
            values. Will default to returning the invocation info on all
        """
        log.info("Beginning remote Pipeline invocation")
        valid_states = [
            PSInvocationState.NOT_STARTED,
            PSInvocationState.STOPPED,
            PSInvocationState.COMPLETED,
            PSInvocationState.FAILED,
        ]
        if self.state not in valid_states:
            raise InvalidPipelineStateError(self.state, valid_states, "start a PowerShell pipeline")

        if len(self.commands) == 0:
            raise InvalidPSRPOperation("Cannot invoke PowerShell without any commands being set")

        no_input = input is None or not input
        apartment_state = apartment_state or self.runspace_pool.apartment_state
        host_info = HostInfo(host=self.runspace_pool.host)

        pipeline = Pipeline(
            is_nested=self.is_nested,
            cmds=self.commands,
            history=self.history_string,
            redirect_err_to_out=redirect_shell_error_to_out,
        )
        create_pipeline = CreatePipeline(
            no_input,
            ApartmentState(value=apartment_state),
            RemoteStreamOptions(value=remote_stream_options),
            add_to_history,
            host_info,
            pipeline,
            self.is_nested,
        )
        self._invoke(create_pipeline)

        # finally send the input if any was specified
        if input is not None:
            if not isinstance(input, types.GeneratorType):

                def input_gen(i: bytes) -> typing.Iterator[bytes]:
                    yield i

                input = input_gen(input)  # type: ignore[arg-type]

            log.info("Sending input to remote Pipeline")
            next_input = next(input)
            while next_input is not None:
                if not isinstance(next_input, list):
                    next_input = [next_input]

                input_msgs: typing.List[ComplexObject] = [PipelineInput(data=d) for d in next_input]

                try:
                    next_input = next(input)
                except StopIteration:
                    input_msgs.append(EndOfPipelineInput())
                    next_input = None

                fragments = self.runspace_pool._fragmenter.fragment_multiple(input_msgs, self.runspace_pool.id, self.id)

                for fragment in fragments:
                    self.runspace_pool.shell.send("stdin", fragment, command_id=self._command_id)

    def end_invoke(self) -> typing.List:
        """
        Wait until the asynchronous command has finished executing and return
        the output collection of return objects.

        :return: A list of output objects
        """
        while self.state == PSInvocationState.RUNNING:
            self.poll_invoke()

        return self.output

    def invoke(
        self,
        input: typing.Optional[typing.Union[bytes, typing.Iterator]] = None,
        add_to_history: bool = False,
        apartment_state: typing.Optional[int] = None,
        redirect_shell_error_to_out: bool = False,
        remote_stream_options: typing.Optional[int] = RemoteStreamOptions.ADD_INVOCATION_INFO,
    ) -> typing.List:
        """
        Invoke the command and return the output collection of return objects.

        :param input: List of inputs to the command, this will be manually
            serialized but can also be a result of runspace_pool.serialize()
        :param add_to_history: Add the commands run to the pool history
        :param apartment_state: Override the RunspacePool apartment state with
            one just for this invocation
        :param redirect_shell_error_to_out: Whether to redirect the global
            error output pipe to the commands error output pipe.
        :param remote_stream_options: Whether to return the invocation info on
            the various steams, see complex_objects.RemoteStreamOptions for the
            values. Will default to returning the invocation info on all
        :return: A list of output objects
        """
        self.begin_invoke(input, add_to_history, apartment_state, redirect_shell_error_to_out, remote_stream_options)
        return self.end_invoke()

    def merge_previous(
        self,
        enabled: bool = False,
    ) -> "PowerShell":
        """
        Sets the MergePreviousResults of the last command in the pipeline.
        This is the same as the MergeUnclaimedPreviousCommandResults property
        used in the .NET System.Management.Automation.Runspaces.Command class.

        :param enabled: Whether to merge previous results to the Output and
            Error streams or not
        :return: The current PowerShell instance with the last command set to
            MergePreviousResults
        """
        if enabled:
            value = PipelineResultTypes.OUTPUT | PipelineResultTypes.ERROR
        else:
            value = PipelineResultTypes.NONE
        pipeline = PipelineResultTypes(protocol_version_2=True, value=value)
        self.commands[-1].merge_previous = pipeline
        return self

    def merge_all(
        self,
        to: str = "none",
    ) -> "PowerShell":
        """
        Will merge all relevant streams to the stream specified of the last
        command in the pipeline.

        :param to: Can be one of the following
            none: Do not merge the streams
            output: Send all streams to the Output stream
        :return: The current PowerSHell instance with the last command set to
            Merge<stream> to the stream desired.
        """
        self.merge_error(to)

        if version_equal_or_newer(self.runspace_pool.protocol_version or "", "2.2"):
            self.merge_debug(to)
            self.merge_verbose(to)
            self.merge_warning(to)

        if version_equal_or_newer(self.runspace_pool.protocol_version or "", "2.3"):
            self.merge_information(to)
        return self

    def merge_error(
        self,
        to: str = "none",
    ) -> "PowerShell":
        """
        Controls where the Error stream of the last command in the pipeline
        will be sent to.

        :param to: Can be one of the following
            none: Send all Error streams to the Error stream
            output: Send all Error streams to the Output stream
        :return: The current PowerShell instance with the last command set to
            MergeError to the stream desired.
        """
        self._set_merge_to("merge_error", to, ["none", "output"])

        # For V2 backwards compatibility
        if to == "none":
            my_result = self.commands[-1].merge_error
        else:
            my_result = PipelineResultTypes(value=PipelineResultTypes.ERROR)
        self.commands[-1].merge_my_result = my_result
        self.commands[-1].merge_to_result = self.commands[-1].merge_error
        return self

    def merge_warning(
        self,
        to: str = "none",
    ) -> "PowerShell":
        """
        Controls where the Warning stream of the last command in the pipeline
        will be sent to. This will throw an InvalidPSRPOperation error if the
        protocol version of the server is less than 2.2 (PSv2).

        :param to: Can be on of the following
            none: Send all Warning streams to the Warning stream
            output: Send all Warning streams to the Output stream
            null: Sends it to a null pipe (this doesn't seem to do anything)
        :return: The current PowerShell instance with the last command set to
            MergeWarning to the stream desired.
        """
        self._set_merge_to("merge_warning", to, None, "2.2")
        return self

    def merge_verbose(
        self,
        to: str = "none",
    ) -> "PowerShell":
        """
        Controls where the Verbose stream of the last command in the pipeline
        will be sent to. This will throw an InvalidPSRPOperation error if the
        protocol version of the server is less than 2.2 (PSv2).

        :param to: Can be on of the following
            none: Send all Verbose streams to the Verbose stream
            output: Send all Verbose streams to the Output stream
            null: Sends it to a null pipe (this doesn't seem to do anything)
        :return: The current PowerShell instance with the last command set to
            MergeVerbose to the stream desired.
        """
        self._set_merge_to("merge_verbose", to, None, "2.2")
        return self

    def merge_debug(
        self,
        to: str = "none",
    ) -> "PowerShell":
        """
        Controls where the Debug stream of the last command in the pipeline
        will be sent to. This will throw an InvalidPSRPOperation error if the
        protocol version of the server is less than 2.2 (PSv2).

        :param to: Can be on of the following
            none: Send all Debug streams to the Debug stream
            output: Send all Debug streams to the Output stream
            null: Sends it to a null pipe (this doesn't seem to do anything)
        :return: The current PowerShell instance with the last command set to
            MergeDebug to the stream desired.
        """
        self._set_merge_to("merge_debug", to, None, "2.2")
        return self

    def merge_information(
        self,
        to: str = "none",
    ) -> "PowerShell":
        """
        Controls where the Information stream of the last command in the
        pipeline will be sent to. This will throw an InvalidPSRPOperation error
        if the protocol version of the server is less than 2.3 (PSv5).

        :param to: Can be on of the following
            none: Send all Information streams to the Information stream
            output: Send all Information streams to the Output stream
            null: Sends it to a null pipe (this doesn't seem to do anything)
        :return: The current PowerShell instance with the last command set to
            MergeInformation to the stream desired.
        """
        self._set_merge_to("merge_information", to, None, "2.3")
        return self

    def merge_reset(self) -> "PowerShell":
        """
        Will reset the merge behaviour of all streams back to the default. This
        is done on the last command in the pipeline.

        :return: The current PowerShell instance with the last commands merge
            behaviour set to the default.
        """
        self.merge_previous(enabled=False)
        self.merge_all("none")
        return self

    def poll_invoke(
        self,
        timeout: typing.Optional[int] = None,
    ) -> None:
        """
        Poll the running process to update the output streams and the status.

        :param timeout: Override the default WSMan timeout when polling the
        pipeline.
        """
        try:
            responses = self.runspace_pool._receive(self.id, timeout=timeout)
        except WSManFaultError as err:
            # operation timeout needs to be ignored and silently tried
            # again
            if err.code == 2150858793:
                responses = []
            else:
                raise err

        for response in responses:
            if response[0] == MessageType.PIPELINE_OUTPUT:
                self.output.append(response[1].data.data)

    def stop(self) -> None:
        """
        Stop the currently running command.
        """
        log.info("Stopping remote Pipeline")
        if self.state in [PSInvocationState.STOPPING, PSInvocationState.STOPPED]:
            return
        elif self.state != PSInvocationState.RUNNING:
            raise InvalidPipelineStateError(self.state, PSInvocationState.RUNNING, "stop a running pipeline")

        self.state = PSInvocationState.STOPPING
        self.runspace_pool.shell.signal(SignalCode.PS_CTRL_C, str(self.id).upper())
        self.state = PSInvocationState.STOPPED
        self.runspace_pool.pipelines.pop(self.id, None)

    def _invoke(
        self,
        msg: ComplexObject,
    ) -> None:
        fragments = self.runspace_pool._fragmenter.fragment(msg, self.runspace_pool.id, self.id)

        self.runspace_pool.pipelines[self.id] = self

        # send first fragment as Command message
        first_frag = base64.b64encode(fragments.pop(0)).decode("utf-8")
        resp = self.runspace_pool.shell.command("", arguments=[first_frag], command_id=self.id)
        cmd_id = resp.find("rsp:CommandResponse/rsp:CommandId", NAMESPACES)
        if cmd_id is not None:
            self._command_id = cmd_id.text or ""
        self.state = PSInvocationState.RUNNING

        # send the remaining fragments with the Send message
        for fragment in fragments:
            self.runspace_pool.shell.send("stdin", fragment, command_id=self._command_id)

    def _set_merge_to(
        self,
        merge: str,
        to: str,
        valid: typing.Optional[typing.List[str]],
        min_protocol: typing.Optional[str] = None,
    ) -> None:
        valid = valid or ["none", "null", "output"]
        if to not in valid:
            raise InvalidPSRPOperation("Invalid merge to option '%s', valid values %s" % (to, ", ".join(valid)))

        if min_protocol is not None:
            if not version_equal_or_newer(self.runspace_pool.protocol_version or "", min_protocol):
                error_msg = (
                    "Merge option for '%s' is not supported in the current "
                    "protocol version %s, minimum version required %s"
                    % (merge, self.runspace_pool.protocol_version, min_protocol)
                )
                raise InvalidPSRPOperation(error_msg)

        to_value = getattr(PipelineResultTypes, to.upper())
        to_result = PipelineResultTypes(value=to_value)
        setattr(self.commands[-1], merge, to_result)

    def _process_error_record(
        self,
        message: Message,
    ) -> None:
        self.streams.error.append(typing.cast(ErrorRecordMessage, message.data))

    def _process_pipeline_host_call(
        self,
        message: Message,
    ) -> ComplexObject:
        return self.runspace_pool._process_host_call(message, PipelineHostResponse, self)

    def _process_pipeline_state(
        self,
        message: Message,
    ) -> None:
        msg_state = typing.cast(PipelineState, message.data)

        log.debug("Received PipelineState with state: %d" % msg_state.state)
        self.state = msg_state.state
        if msg_state.error_record is not None:
            self.streams.error.append(msg_state.error_record)

        if self.state == PSInvocationState.FAILED:
            self.had_errors = True

    def _process_debug_record(
        self,
        message: Message,
    ) -> None:
        self.streams.debug.append(typing.cast(DebugRecord, message.data))

    def _process_verbose_record(
        self,
        message: Message,
    ) -> None:
        self.streams.verbose.append(typing.cast(VerboseRecord, message.data))

    def _process_warning_record(
        self,
        message: Message,
    ) -> None:
        self.streams.warning.append(typing.cast(WarningRecord, message.data))

    def _process_progress_record(
        self,
        message: Message,
    ) -> None:
        self.streams.progress.append(typing.cast(ProgressRecord, message.data))

    def _process_information_record(
        self,
        message: Message,
    ) -> None:
        self.streams.information.append(typing.cast(InformationRecord, message.data))


class PSDataStreams(object):
    def __init__(self) -> None:
        """
        Streams generated by PowerShell invocations

        System.Management.Automation.PSDataStreams
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.psdatastreams?view=powershellsdk-1.1.0
        """
        self.debug: typing.List[DebugRecord] = []
        self.error: typing.List[ErrorRecordMessage] = []
        self.information: typing.List[InformationRecord] = []
        self.progress: typing.List[ProgressRecord] = []
        self.verbose: typing.List[VerboseRecord] = []
        self.warning: typing.List[WarningRecord] = []


class Fragmenter(object):
    def __init__(
        self,
        max_size: int,
        serializer: Serializer,
    ) -> None:
        self.incoming_buffer: typing.Dict[int, typing.Dict] = {}
        self.outgoing_counter = 1
        self.max_size = max_size - 21  # take away the fragment header size
        self.serializer = serializer

    def fragment(
        self,
        data: ComplexObject,
        rpid: str,
        pid: typing.Optional[str] = None,
        remaining_size: typing.Optional[int] = None,
    ) -> typing.List[bytes]:
        msg = Message(Destination.SERVER, rpid, pid, data, self.serializer)
        msg_data = msg.pack()
        max_size = self.max_size
        start = True
        fragment_id = 0
        fragments: typing.List[bytes] = []

        if remaining_size is not None:
            msg_fragment = msg_data[:remaining_size]
            msg_data = msg_data[len(msg_fragment) :]

            end = msg_data == b""
            fragment = Fragment(self.outgoing_counter, fragment_id, msg_fragment, start, end)
            fragments.append(fragment.pack())
            fragment_id += 1
            start = False

        for msg_fragment, end in self._byte_iterator(msg_data, max_size):
            fragment = Fragment(self.outgoing_counter, fragment_id, msg_fragment, start, end)
            fragments.append(fragment.pack())
            fragment_id += 1
            start = False

        self.outgoing_counter += 1
        return fragments

    def fragment_multiple(
        self,
        blocks: typing.List[ComplexObject],
        rpid: str,
        pid: typing.Optional[str] = None,
    ) -> typing.List[bytes]:
        remaining_size = self.max_size
        fragments: typing.List[bytes] = []

        for block in blocks:
            block_fragments = self.fragment(block, rpid, pid, remaining_size=remaining_size)

            # the first fragment can fit in with the previous block fragment
            # append to the last fragment and remove from the list
            if remaining_size != self.max_size and len(fragments) > 0:
                fragments[-1] = fragments[-1] + block_fragments.pop(0)

            fragments.extend(block_fragments)

            # calculate how much data can fit into the last fragment
            remaining_size = self.max_size - len(fragments[-1])
            if remaining_size <= 0:
                remaining_size = self.max_size

        return fragments

    def defragment(
        self,
        data: bytes,
    ) -> typing.List[Message]:
        fragments = []
        while data != b"":
            frag, data = Fragment.unpack(data)
            incoming_buffer = self.incoming_buffer.get(frag.object_id)
            if incoming_buffer is None:
                incoming_buffer = {"data": b"", "id": 0}
                self.incoming_buffer[frag.object_id] = incoming_buffer

            if frag.fragment_id != incoming_buffer["id"]:
                raise FragmentError(
                    "Fragment Fragment Id: %d != Expected Fragment Id: %d" % (frag.fragment_id, incoming_buffer["id"])
                )

            if frag.start and frag.end:
                fragments.append(frag.data)
                del self.incoming_buffer[frag.object_id]
            elif frag.start:
                incoming_buffer["data"] = frag.data
                incoming_buffer["id"] += 1
            elif frag.end:
                fragments.append(incoming_buffer["data"] + frag.data)
                del self.incoming_buffer[frag.object_id]
            else:
                incoming_buffer["data"] += frag.data
                incoming_buffer["id"] += 1

        messages = [Message.unpack(fragment, self.serializer) for fragment in fragments]
        return messages

    def _byte_iterator(
        self,
        data: bytes,
        buffer_size: int,
    ) -> typing.Iterator[typing.Tuple[bytes, bool]]:
        byte_count = len(data)
        for i in range(0, byte_count, buffer_size):
            yield data[i : i + buffer_size], i + buffer_size >= byte_count


class Fragment(object):
    def __init__(
        self,
        object_id: int,
        fragment_id: int,
        data: bytes,
        start: bool = False,
        end: bool = False,
    ) -> None:
        self.object_id = object_id
        self.fragment_id = fragment_id
        self.start = start
        self.end = end
        self.data = data

    def pack(self) -> bytes:
        start_end_byte = 0
        if self.start:
            start_end_byte |= 0x1
        if self.end:
            start_end_byte |= 0x2

        data = struct.pack(">Q", self.object_id)
        data += struct.pack(">Q", self.fragment_id)
        data += struct.pack("B", start_end_byte)
        data += struct.pack(">I", len(self.data))
        data += self.data

        return data

    @staticmethod
    def unpack(data: bytes) -> typing.Tuple["Fragment", bytes]:
        object_id = struct.unpack(">Q", data[0:8])[0]
        fragment_id = struct.unpack(">Q", data[8:16])[0]

        start_end_byte = struct.unpack("B", data[16:17])[0]
        start = start_end_byte & 0x1 == 0x1
        end = start_end_byte & 0x2 == 0x2

        length = struct.unpack(">I", data[17:21])[0]
        fragment_data = data[21 : length + 21]

        fragment = Fragment(object_id, fragment_id, fragment_data, start, end)
        return fragment, data[21 + length :]



import base64
import os
import re
import struct
import time
import uuid
import xml.etree.ElementTree as ET

import pytest
import requests.exceptions
import yaml

from pypsrp._utils import to_bytes, to_string
from pypsrp.exceptions import AuthenticationError, WinRMTransportError
from pypsrp.powershell import Fragment
from pypsrp.wsman import NAMESPACES, WSMan

from . import assert_xml_diff


class TransportFake(object):
    def __init__(self, test_name, server, port, username, password, ssl, path, auth):
        """
        This is a fake transport stub that takes in known requests and responds
        back with the already created response. This is used in cases when a
        real WinRM endpoint is not available or a fake one was requested for
        a predefined apth

        :param test_name: The name of the test which is used to get the
            messages to send to the server
        """
        self.server = server
        self.port = port
        self.username = username
        self.password = password
        self.ssl = ssl
        self.path = path
        self.auth = auth
        self.endpoint = "%s://%s:%d/%s" % ("https" if ssl else "http", server, port, path)
        self.session = None

        # used in the test only
        for key, value in NAMESPACES.items():
            ET.register_namespace(key, value)

        self._uuid_pattern = re.compile("[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", re.I)
        self._test_name = test_name
        self._msg_counter = 0
        self._psrp_fragments = {}  # used to store PSRP fragments as they come in
        meta_path = os.path.join(os.path.dirname(__file__), "responses/%s.yml" % test_name)

        if os.path.exists(meta_path):
            with open(meta_path, "rb") as o:
                self._test_meta = yaml.load(o, Loader=yaml.SafeLoader)
        else:
            raise Exception("Test metadata yml file does not exist at %s" % meta_path)

        self._test_msg_key = "messages"

    def close(self):
        return

    def send(self, message, *args, **kwargs):
        current_msg = self._test_meta[self._test_msg_key][self._msg_counter]
        actual = self._normalise_xml(message, generify=False, psrp_fragment_type="actual")
        expected = self._normalise_xml(
            current_msg["request"], overrides=current_msg.get("overrides", None), psrp_fragment_type="expected"
        )
        failure_msg = "Message %d request for test %s does not match expectation\nActual:   %s\nExpected: %s" % (
            self._msg_counter,
            self._test_name,
            actual,
            expected,
        )
        assert_xml_diff(actual, expected, msg=failure_msg)

        for obj_id in list(self._psrp_fragments.keys()):
            details = self._psrp_fragments[obj_id]
            if not details["end"]:
                continue

            fragment_ids = list(details["actual"].keys())
            fragment_ids.sort()

            actual_obj = b""
            expected_obj = b""
            for i in range(0, fragment_ids[-1] + 1):
                actual_obj += details["actual"][i]
                expected_obj += details["expected"][i]

            actual_destination = struct.unpack("<I", actual_obj[:4])[0]
            actual_message_type = struct.unpack("<I", actual_obj[4:8])[0]
            actual_message = self._normalise_xml(actual_obj[40:], generify=False)
            expected_destination = struct.unpack("<I", expected_obj[:4])[0]
            expected_message_type = struct.unpack("<I", expected_obj[4:8])[0]
            expected_message = self._normalise_xml(expected_obj[40:], generify=False)

            assert actual_destination == expected_destination
            assert actual_message_type == expected_message_type

            # Only do the XML compare if the message had data, otherwise just
            # compare the actual values
            failure_msg = (
                "PSRP Message object %d for test %s does not match "
                "expectation\nActual:    %s\nExpected: %s" % (obj_id, self._test_name, actual_message, expected_message)
            )
            assert_xml_diff(actual_message, expected_message, msg=failure_msg)

            # Remove the fragments for the obj as we've verified them
            del self._psrp_fragments[obj_id]

        response = self._normalise_xml(current_msg["response"])
        self._msg_counter += 1

        # check if test metadata indicates we want to raise an exception here
        # instead of returning the response
        if "transport_error" in current_msg.keys():
            raise WinRMTransportError(
                current_msg["transport_error"]["protocol"], current_msg["transport_error"]["code"], response
            )
        elif current_msg.get("auth_error", False):
            raise AuthenticationError("Failed to authenticate the user %s with %s" % (self.username, self.auth))
        elif current_msg.get("http_error", False):
            raise requests.exceptions.ConnectionError()

        if "timeout" in current_msg.keys():
            time.sleep(current_msg["timeout"])

        return response

    def _normalise_xml(self, xml, generify=True, overrides=None, psrp_fragment_type=None):
        if not xml:
            return xml

        overrides = overrides if overrides is not None else []

        if generify:
            # convert all UUID values to the blank UUID
            xml = re.sub(self._uuid_pattern, "00000000-0000-0000-0000-000000000000", xml)

            xml_obj = ET.fromstring(to_bytes(xml))

            # convert the To hostname in the headers to the generic one
            to_field = xml_obj.find("s:Header/wsa:To", NAMESPACES)
            if to_field is not None:
                to_field.text = self.endpoint

            for override in overrides:
                override_element = xml_obj.find(override["path"], NAMESPACES)
                if override.get("text"):
                    override_element.text = override["text"]
                attributes = override.get("attributes", {})
                for attr_key, attr_value in attributes.items():
                    override_element.attrib[attr_key] = attr_value
        else:
            xml_obj = ET.fromstring(to_bytes(xml))

        # PSRP message contain another set of XML messages that have been
        # base64 encoded. We need to strip these out and compare them
        # separately once all the fragments have been received.
        if psrp_fragment_type:
            creation_xml = xml_obj.find(
                "s:Body/rsp:Shell/{http://schemas.microsoft.com/powershell}creationXml", NAMESPACES
            )
            if creation_xml is not None:
                creation_xml.text = self._generify_fragment(creation_xml.text, psrp_fragment_type)

            connect_xml = xml_obj.find("s:Body/rsp:Connect/pwsh:connectXml", NAMESPACES)
            if connect_xml is not None:
                connect_xml.text = self._generify_fragment(connect_xml.text, psrp_fragment_type)

            # when resource uri is PowerShell we know the Send/Command messages
            # contain PSRP fragments and we need to generify them
            exp_res_uri = "http://schemas.microsoft.com/powershell/"
            res_uri = xml_obj.find("s:Header/wsman:ResourceURI", NAMESPACES)
            if res_uri is not None:
                res_uri = res_uri.text

            streams = xml_obj.findall("s:Body/rsp:Send/rsp:Stream", NAMESPACES)
            if res_uri is not None and res_uri.startswith(exp_res_uri) and len(streams) > 0:
                for stream in streams:
                    stream.text = self._generify_fragment(stream.text, psrp_fragment_type)

            command = xml_obj.find("s:Body/rsp:CommandLine/rsp:Arguments", NAMESPACES)
            if res_uri is not None and res_uri.startswith(exp_res_uri) and command is not None:
                command.text = self._generify_fragment(command.text, psrp_fragment_type)

        return to_string(ET.tostring(xml_obj, encoding="utf-8"))

    def _simplify_namespaces(self, namespaces, element):
        namespaces.update(element.nsmap)

        tag_details = element.tag.split("}", 1)
        if len(tag_details) < 2:
            new_tag = element.tag
        else:
            ns = tag_details[0][1:]
            namespace_tag = namespaces.keys()[namespaces.values().index(ns)]
            new_tag = "%s:%s" % (namespace_tag, tag_details[1])

        new_attributes = {}
        for key, value in element.attrib.items():
            new_attributes[key] = value

        new_element = ET.Element(new_tag, attrib=new_attributes)
        new_element.text = element.text

        for child_element in element:
            new_child = self._simplify_namespaces(namespaces, child_element)
            new_element.append(new_child)

        return new_element

    def _generify_fragment(self, fragment, fragment_type):
        f_data = base64.b64decode(fragment)
        new_value = b""
        idx = 0

        while idx < len(f_data):
            frag = Fragment.unpack(f_data[idx:])[0]

            length = struct.unpack(">I", f_data[idx + 17 : idx + 21])[0]
            m_data = f_data[idx + 21 : idx + length + 21]

            # We store the fragments for later comparison
            fragment_buffer = self._psrp_fragments.get(frag.object_id, None)
            if fragment_buffer is None:
                fragment_buffer = {"actual": {}, "expected": {}, "end": False}
                self._psrp_fragments[frag.object_id] = fragment_buffer

            fragment_buffer[fragment_type][frag.fragment_id] = m_data
            if frag.end:
                fragment_buffer["end"] = True

            # We don't add the actual message to make the initial message
            # comparison easier
            new_value += f_data[idx : idx + 21]
            idx += length + 21
        return base64.b64encode(new_value).decode("utf-8")


@pytest.fixture(scope="function")
def wsman_conn(request, monkeypatch):
    test_params = request.param
    if not isinstance(test_params, list) or len(test_params) != 2:
        raise Exception("Cannot run winrm_transport fixture without the allow real and test name set")

    allow_real = test_params[0]
    test_name = test_params[1]

    # these need to be set to run against a proper server
    username = os.environ.get("PYPSRP_USERNAME", None)
    password = os.environ.get("PYPSRP_PASSWORD", None)
    server = os.environ.get("PYPSRP_SERVER", None)

    # these are optional vars that can further control the transport setup
    auth = os.environ.get("PYPSRP_AUTH", "negotiate")
    port = int(os.environ.get("PYPSRP_PORT", "5986"))
    ssl = port != 5985

    if allow_real and username is not None and password is not None and server is not None:
        wsman = WSMan(
            server,
            port=port,
            username=username,
            password=password,
            ssl=ssl,
            auth=auth,
            cert_validation=False,
        )
    else:
        # Mock out UUID's so they are not a problem when comparing messages
        def mockuuid():
            return uuid.UUID("00000000-0000-0000-0000-000000000000")

        monkeypatch.setattr(uuid, "uuid4", mockuuid)
        transport = TransportFake(test_name, "fakehost", port, "username", "password", ssl, "wsman", auth)
        wsman = WSMan("")
        wsman.transport = transport

    with wsman:
        yield wsman

    # used as an easy way to be results for a test, requires the _test_messages
    # to be uncommented in pypsrp/wsman.py
    test_messages = getattr(wsman.transport, "_test_messages", None)
    if test_messages is not None:
        yaml_text = "\n\n" + yaml.dump({"messages": test_messages}, default_flow_style=False, width=9999)
        print(yaml_text)



import pytest

from pypsrp._utils import (
    get_hostname,
    to_bytes,
    to_string,
    to_unicode,
    version_equal_or_newer,
)


def test_unicode_to_bytes_default():
    expected = b"\x61\x62\x63"
    actual = to_bytes("abc")
    assert actual == expected


def test_unicode_to_bytes_diff_encoding():
    expected = b"\x61\x00\x62\x00\x63\x00"
    actual = to_bytes("abc", encoding="utf-16-le")
    assert actual == expected


def test_bytes_to_bytes():
    expected = b"\x01\x02\x03\x04"
    actual = to_bytes(b"\x01\x02\x03\x04")
    assert actual == expected


def test_str_to_bytes():
    # Python 3 the default string type is unicode so the expected value will
    # be "abc" in UTF-16 form while Python 2 "abc" is the bytes representation
    # already
    expected = b"\x61\x00\x62\x00\x63\x00"
    actual = to_bytes("abc", encoding="utf-16-le")
    assert actual == expected


def test_unicode_to_unicode():
    expected = "abc"
    actual = to_unicode("abc")
    assert actual == expected


def test_byte_to_unicode():
    expected = "abc"
    actual = to_unicode(b"\x61\x62\x63")
    assert actual == expected


def test_byte_to_unicode_diff_encoding():
    expected = "abc"
    actual = to_unicode(b"\x61\x00\x62\x00\x63\x00", encoding="utf-16-le")
    assert actual == expected


def test_str_to_unicode():
    expected = "a\x00b\x00c\x00"
    actual = to_unicode("a\x00b\x00c\x00", encoding="utf-16-le")
    assert actual == expected


def test_to_str():
    assert str(to_string).startswith("<function to_unicode")


@pytest.mark.parametrize(
    "version, reference_version, expected",
    [
        ["2.2", "2.3", False],
        ["2.3", "2.3", True],
        ["2.4", "2.3", True],
        ["3", "2.3", True],
        ["3.0", "2.3", True],
        ["1", "2.3", False],
        ["1.0", "2.3", False],
        ["2.3.0", "2.3", True],
        ["2.3.1", "2.3", True],
        ["2.3", "2.3.0", True],
        ["2.3", "2.3.1", False],
    ],
)
def test_version_newer(version, reference_version, expected):
    assert version_equal_or_newer(version, reference_version) == expected


@pytest.mark.parametrize(
    "url, expected",
    [
        # hostname
        ["http://hostname", "hostname"],
        ["https://hostname", "hostname"],
        ["http://hostname:1234", "hostname"],
        ["https://hostname:1234", "hostname"],
        ["http://hostname/path", "hostname"],
        ["https://hostname/path", "hostname"],
        ["http://hostname:1234/path", "hostname"],
        ["https://hostname:1234/path", "hostname"],
        # fqdn
        ["http://hostname.domain.com", "hostname.domain.com"],
        ["https://hostname.domain.com", "hostname.domain.com"],
        ["http://hostname.domain.com:1234", "hostname.domain.com"],
        ["https://hostname.domain.com:1234", "hostname.domain.com"],
        ["http://hostname.domain.com/path", "hostname.domain.com"],
        ["https://hostname.domain.com/path", "hostname.domain.com"],
        ["http://hostname.domain.com:1234/path", "hostname.domain.com"],
        ["https://hostname.domain.com:1234/path", "hostname.domain.com"],
        # ip address
        ["http://1.2.3.4", "1.2.3.4"],
        ["https://1.2.3.4", "1.2.3.4"],
        ["http://1.2.3.4:1234", "1.2.3.4"],
        ["https://1.2.3.4:1234", "1.2.3.4"],
        ["http://1.2.3.4/path", "1.2.3.4"],
        ["https://1.2.3.4/path", "1.2.3.4"],
        ["http://1.2.3.4:1234/path", "1.2.3.4"],
        ["https://1.2.3.4:1234/path", "1.2.3.4"],
    ],
)
def test_get_hostname(url, expected):
    assert expected == get_hostname(url)



# Copyright: (c) 2026, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import annotations

import contextvars
import functools
import types
import typing as t

import requests.adapters
from requests.packages.urllib3.util.retry import Retry

T = t.TypeVar("T")


class NewConnectionDisabled(Exception):
    """Raised when new connections are being made but have been disabled in the current context."""


class DisableNewConnectionsContext:
    """Context manager to disable new connections from being created."""

    __slots__ = "_contextvar_token"

    _contextvar: t.ClassVar[contextvars.ContextVar] = contextvars.ContextVar("NewConnectionContext")
    _contextvar_token: contextvars.Token

    @classmethod
    def current(cls) -> DisableNewConnectionsContext | None:
        """Returns the current context or None if not set."""
        try:
            return cls._contextvar.get()
        except LookupError:
            return None

    def __enter__(self) -> DisableNewConnectionsContext:
        self._contextvar_token = self.__class__._contextvar.set(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> bool | None:
        self.__class__._contextvar.reset(self._contextvar_token)
        del self._contextvar_token

        return None


class HTTPSAdapterWithKeyPassword(requests.adapters.HTTPAdapter):

    def __init__(
        self,
        *args: t.Any,
        _pypsrp_key_password: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        self.__key_password = _pypsrp_key_password
        super().__init__(*args, **kwargs)

    def init_poolmanager(
        self,
        connections,
        maxsize,
        block=False,
        **pool_kwargs,
    ):
        return super().init_poolmanager(
            connections,
            maxsize,
            block,
            key_password=self.__key_password,
            **pool_kwargs,
        )


def _wrap(
    func: t.Callable[..., T],
    *,
    before: t.Callable[[], None] | None = None,
    after: t.Callable[[T], None] | None = None,
) -> t.Callable[..., T]:
    """Wraps a function with before and after callables."""

    @functools.wraps(func)
    def wrapper(*args: t.Any, **kwargs: t.Any) -> T:
        if before:
            before()

        res = func(*args, **kwargs)

        if after:
            after(res)

        return res

    return wrapper


def _raise_if_connections_disabled():
    if DisableNewConnectionsContext.current() is not None:
        raise NewConnectionDisabled()


def _create_new_connection_with_failure(connection):
    connection.connect = _wrap(connection.connect, before=_raise_if_connections_disabled)


def _create_connection_pool(pool):
    # This is called in the urllib3 pool when a new connection is needed. We
    # wrap it so we can wrap the connect method when a new connection is
    # created.
    pool.ConnectionCls = _wrap(pool.ConnectionCls, after=_create_new_connection_with_failure)


def create_request_adapter(  # type: ignore[no-any-unimported]  # requests does not have typing stubs for urllib3
    *,
    max_retries: Retry,
    key_password: str | None = None,
) -> requests.adapters.HTTPAdapter:
    """Creates a HTTPAdapter with support for disabling new connections via context."""
    adapter: requests.adapters.HTTPAdapter
    if key_password is not None:
        adapter = HTTPSAdapterWithKeyPassword(
            max_retries=max_retries,
            _pypsrp_key_password=key_password,
        )
    else:
        adapter = requests.adapters.HTTPAdapter(max_retries=max_retries)

    # pool_classes_by_scheme stores the urllib3 pool types used when creating
    # the connection pool for http/https. We wrap the __init__ method so we can
    # inject our custom ConnectionCls that wraps the connect method when
    # created.
    pool_classes = adapter.poolmanager.pool_classes_by_scheme
    for scheme, pool_cls in list(pool_classes.items()):
        pool_classes[scheme] = _wrap(pool_cls, after=_create_connection_pool)

    return adapter



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing
from copy import deepcopy

from pypsrp._utils import to_string, version_equal_or_newer


class ObjectMeta(object):
    def __init__(
        self,
        tag: str = "*",
        name: typing.Optional[str] = None,
        optional: bool = False,
        object: typing.Any = None,
    ) -> None:
        self.tag = tag
        self.name = name
        self.optional = optional
        self.object = object


class ListMeta(ObjectMeta):
    def __init__(
        self,
        obj_type: str = "LST",
        name: typing.Optional[str] = None,
        optional: bool = False,
        list_value_meta: typing.Optional[ObjectMeta] = None,
        list_types: typing.Optional[typing.List[str]] = None,
    ) -> None:
        super(ListMeta, self).__init__(obj_type, name, optional)

        if list_value_meta is None:
            self.list_value_meta = ObjectMeta()
        else:
            self.list_value_meta = list_value_meta

        if list_types is None:
            self.list_types = ["System.Object[]", "System.Array", "System.Object"]
        else:
            self.list_types = list_types


class StackMeta(ListMeta):
    def __init__(
        self,
        name: typing.Optional[str] = None,
        optional: bool = False,
        list_value_meta: typing.Optional[ObjectMeta] = None,
        list_types: typing.Optional[typing.List[str]] = None,
    ) -> None:
        if list_types is None:
            list_types = ["System.Collections.Stack", "System.Object"]
        super(StackMeta, self).__init__("STK", name, optional, list_value_meta, list_types)


class QueueMeta(ListMeta):
    def __init__(
        self,
        name: typing.Optional[str] = None,
        optional: bool = False,
        list_value_meta: typing.Optional[ObjectMeta] = None,
        list_types: typing.Optional[typing.List[str]] = None,
    ) -> None:
        if list_types is None:
            list_types = ["System.Collections.Queue", "System.Object"]
        super(QueueMeta, self).__init__("QUE", name, optional, list_value_meta, list_types)


class DictionaryMeta(ObjectMeta):
    def __init__(
        self,
        name: typing.Optional[str] = None,
        optional: bool = False,
        dict_key_meta: typing.Optional[ObjectMeta] = None,
        dict_value_meta: typing.Optional[ObjectMeta] = None,
        dict_types: typing.Optional[typing.List[str]] = None,
    ) -> None:
        super(DictionaryMeta, self).__init__("DCT", name, optional)
        if dict_key_meta is None:
            self.dict_key_meta = ObjectMeta(name="Key")
        else:
            self.dict_key_meta = dict_key_meta

        if dict_value_meta is None:
            self.dict_value_meta = ObjectMeta(name="Value")
        else:
            self.dict_value_meta = dict_value_meta

        if dict_types is None:
            self.dict_types = ["System.Collections.Hashtable", "System.Object"]
        else:
            self.dict_types = dict_types


class ComplexObject(object):
    def __init__(self) -> None:
        self._adapted_properties: typing.Tuple[typing.Tuple[str, ObjectMeta], ...] = ()
        self._extended_properties: typing.Tuple[typing.Tuple[str, ObjectMeta], ...] = ()
        self._property_sets: typing.Tuple[typing.Tuple[str, ObjectMeta], ...] = ()
        self._types: typing.List[str] = []
        self._to_string = None
        self._xml: typing.Optional[str] = None  # only populated on deserialization

    def __str__(self) -> str:
        return to_string(self._to_string)


class GenericComplexObject(ComplexObject):
    def __init__(self) -> None:
        super(GenericComplexObject, self).__init__()
        self.property_sets: typing.List[typing.Any] = []
        self.extended_properties: typing.Dict[str, typing.Any] = {}
        self.adapted_properties: typing.Dict[str, typing.Any] = {}
        self.to_string: typing.Optional[str] = None
        self.types: typing.List[str] = []

    def __str__(self) -> str:
        return to_string(self.to_string)


class Enum(ComplexObject):
    def __init__(
        self,
        enum_type: typing.Optional[str],
        string_map: typing.Dict[int, str],
        **kwargs: typing.Any,
    ) -> None:
        super(Enum, self).__init__()
        self._types = ["System.Enum", "System.ValueType", "System.Object"]
        if enum_type is not None:
            self._types.insert(0, enum_type)

        self._property_sets = (("value", ObjectMeta("I32")),)
        self._string_map = string_map

        self.value = kwargs.get("value")

    @property  # type: ignore[override]
    def _to_string(self) -> str:
        try:
            return self._string_map[self.value or 0]
        except KeyError as err:
            raise KeyError(
                "%s is not a valid enum value for %s, valid values are %s" % (err, self._types[0], self._string_map)
            )

    @_to_string.setter
    def _to_string(self, value: str) -> None:
        pass


# PSRP Complex Objects - https://msdn.microsoft.com/en-us/library/dd302883.aspx
class Coordinates(ComplexObject):
    def __init__(
        self,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.1 Coordinates
        https://msdn.microsoft.com/en-us/library/dd302883.aspx

        :param x: The X coordinate (0 is the leftmost column)
        :param y: The Y coordinate (0 is the topmost row)
        """
        super(Coordinates, self).__init__()
        self._adapted_properties = (
            ("x", ObjectMeta("I32", name="X")),
            ("y", ObjectMeta("I32", name="Y")),
        )
        self._types = ["System.Management.Automation.Host.Coordinates", "System.ValueType", "System.Object"]
        self.x = kwargs.get("x")
        self.y = kwargs.get("y")


class Size(ComplexObject):
    def __init__(self, **kwargs: typing.Any) -> None:
        """
        [MS-PSRP] 2.2.3.2 Size
        https://msdn.microsoft.com/en-us/library/dd305083.aspx

        :param width: The width of the size
        :param height: The height of the size
        """
        super(Size, self).__init__()
        self._adapted_properties = (
            ("width", ObjectMeta("I32", name="Width")),
            ("height", ObjectMeta("I32", name="Height")),
        )
        self._types = ["System.Management.Automation.Host.Size", "System.ValueType", "System.Object"]
        self.width = kwargs.get("width")
        self.height = kwargs.get("height")


class Color(Enum):
    BLACK = 0
    DARK_BLUE = 1
    DARK_GREEN = 2
    DARK_CYAN = 3
    DARK_RED = 4
    DARK_MAGENTA = 5
    DARK_YELLOW = 6
    GRAY = 7
    DARK_GRAY = 8
    BLUE = 9
    GREEN = 10
    CYAN = 11
    RED = 12
    MAGENTA = 13
    YELLOW = 14
    WHITE = 15

    def __init__(
        self,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.3 Color
        https://msdn.microsoft.com/en-us/library/dd360026.aspx

        :param value: The enum value for Color
        """
        string_map = {
            0: "Black",
            1: "DarkBlue",
            2: "DarkGreen",
            3: "DarkCyan",
            4: "DarkRed",
            5: "DarkMagenta",
            6: "DarkYellow",
            7: "Gray",
            8: "DarkGray",
            9: "Blue",
            10: "Green",
            11: "Cyan",
            12: "Red",
            13: "Magenta",
            14: "Yellow",
            15: "White",
        }
        super(Color, self).__init__("System.ConsoleColor", string_map, **kwargs)


class RunspacePoolState(object):
    BEFORE_OPEN = 0
    OPENING = 1
    OPENED = 2
    CLOSED = 3
    CLOSING = 4
    BROKEN = 5
    NEGOTIATION_SENT = 6
    NEGOTIATION_SUCCEEDED = 7
    CONNECTING = 8
    DISCONNECTED = 9

    def __init__(self, state):
        """
        [MS-PSRP] 2.2.3.4 RunspacePoolState
        https://msdn.microsoft.com/en-us/library/dd341723.aspx

        Represents the state of the RunspacePool.

        :param state: The state int value
        """
        self.state = state

    def __str__(self):
        return {
            0: "BeforeOpen",
            1: "Opening",
            2: "Opened",
            3: "Closed",
            4: "Closing",
            5: "Broken",
            6: "NegotiationSent",
            7: "NegotiationSucceeded",
            8: "Connecting",
            9: "Disconnected",
        }[self.state]


class PSInvocationState(object):
    NOT_STARTED = 0
    RUNNING = 1
    STOPPING = 2
    STOPPED = 3
    COMPLETED = 4
    FAILED = 5
    DISCONNECTED = 6

    def __init__(self, state):
        """
        [MS-PSRP] 2.2.3.5 PSInvocationState
        https://msdn.microsoft.com/en-us/library/dd341651.aspx

        Represents the state of a pipeline invocation.

        :param state: The state int value
        """
        self.state = state

    def __str__(self):
        return {
            0: "NotStarted",
            1: "Running",
            2: "Stopping",
            3: "Stopped",
            4: "Completed",
            5: "Failed",
            6: "Disconnected",
        }[self.state]


class PSThreadOptions(Enum):
    DEFAULT = 0
    USE_NEW_THREAD = 1
    REUSE_THREAD = 2
    USE_CURRENT_THREAD = 3

    def __init__(
        self,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.6 PSThreadOptions
        https://msdn.microsoft.com/en-us/library/dd305678.aspx

        :param value: The enum value for PS Thread Options
        """
        string_map = {0: "Default", 1: "UseNewThread", 2: "ReuseThread", 3: "UseCurrentThread"}
        super(PSThreadOptions, self).__init__(
            "System.Management.Automation.Runspaces.PSThreadOptions", string_map, **kwargs
        )


class ApartmentState(Enum):
    STA = 0
    MTA = 1
    UNKNOWN = 2

    def __init__(
        self,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.7 ApartmentState
        https://msdn.microsoft.com/en-us/library/dd304257.aspx

        :param value: The enum value for Apartment State
        """
        string_map = {0: "STA", 1: "MTA", 2: "UNKNOWN"}
        super(ApartmentState, self).__init__(
            "System.Management.Automation.Runspaces.ApartmentState", string_map, **kwargs
        )


class RemoteStreamOptions(Enum):
    ADD_INVOCATION_INFO_TO_ERROR_RECORD = 1
    ADD_INVOCATION_INFO_TO_WARNING_RECORD = 2
    ADD_INVOCATION_INFO_TO_DEBUG_RECORD = 4
    ADD_INVOCATION_INFO_TO_VERBOSE_RECORD = 8
    ADD_INVOCATION_INFO = 15

    def __init__(
        self,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.8 RemoteStreamOptions
        https://msdn.microsoft.com/en-us/library/dd303829.aspx

        :param value: The initial RemoteStreamOption to set
        """
        super(RemoteStreamOptions, self).__init__(
            "System.Management.Automation.Runspaces.RemoteStreamOptions", {}, **kwargs
        )

    @property  # type: ignore[override]
    def _to_string(self) -> str:
        if self.value == 15:
            return "AddInvocationInfo"

        string_map = (
            ("AddInvocationInfoToErrorRecord", 1),
            ("AddInvocationInfoToWarningRecord", 2),
            ("AddInvocationInfoToDebugRecord", 4),
            ("AddInvocationInfoToVerboseRecord", 8),
        )
        values = []
        for name, flag in string_map:
            if (self.value or 0) & flag == flag:
                values.append(name)
        return ", ".join(values)

    @_to_string.setter
    def _to_string(self, value):
        pass


class Pipeline(ComplexObject):
    class _ExtraCmds(ComplexObject):
        def __init__(self, **kwargs):
            # Used to encapsulate ExtraCmds in the structure required
            super(Pipeline._ExtraCmds, self).__init__()
            self._extended_properties = (
                (
                    "cmds",
                    ListMeta(
                        name="Cmds",
                        list_value_meta=ObjectMeta("Obj", object=Command),
                        list_types=[
                            "System.Collections.Generic.List`1[["
                            "System.Management.Automation.PSObject, "
                            "System.Management.Automation, Version=1.0.0.0, "
                            "Culture=neutral, PublicKeyToken=31bf3856ad364e35]]",
                            "System.Object",
                        ],
                    ),
                ),
            )
            self.cmds = kwargs.get("cmds")

    def __init__(
        self,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.11 Pipeline
        https://msdn.microsoft.com/en-us/library/dd358182.aspx

        :param is_nested: Whether the pipeline is a nested pipeline
        :param commands: List of commands to run
        :param history: The history string to add to the pipeline
        :param redirect_err_to_out: Whether to redirect the global
            error output pipe to the commands error output pipe.
        """
        super(Pipeline, self).__init__()
        cmd_types = [
            "System.Collections.Generic.List`1[["
            "System.Management.Automation.PSObject, "
            "System.Management.Automation, "
            "Version=1.0.0.0, Culture=neutral, "
            "PublicKeyToken=31bf3856ad364e35]]",
            "System.Object",
        ]

        self._extended_properties = (
            ("is_nested", ObjectMeta("B", name="IsNested")),
            # ExtraCmds isn't in spec but is value and used to send multiple
            # statements
            (
                "_extra_cmds",
                ListMeta(
                    name="ExtraCmds", list_value_meta=ObjectMeta("Obj", object=self._ExtraCmds), list_types=cmd_types
                ),
            ),
            ("_cmds", ListMeta(name="Cmds", list_value_meta=ObjectMeta("Obj", object=Command), list_types=cmd_types)),
            ("history", ObjectMeta("S", name="History")),
            ("redirect_err_to_out", ObjectMeta("B", name="RedirectShellErrorOutputPipe")),
        )
        self.is_nested = kwargs.get("is_nested")
        self.commands = kwargs.get("cmds")
        self.history = kwargs.get("history")
        self.redirect_err_to_out = kwargs.get("redirect_err_to_out")

    @property
    def _cmds(self):
        # Cmds is always the first statement
        return self._get_statements()[0]

    @_cmds.setter
    def _cmds(self, value):
        # if commands is already set then that means ExtraCmds was present and
        # has already been set
        if self.commands and len(self.commands) > 0:
            return

        # ExtraCmds wasn't present so we need to unpack it
        self.commands = value

    @property
    def _extra_cmds(self):
        statements = self._get_statements()

        # ExtraCmds is only set if we have more than 1 statement, not present
        # if only 1
        if len(statements) < 2:
            return None
        else:
            extra = [self._ExtraCmds(cmds=c) for c in statements]
            return extra

    @_extra_cmds.setter
    def _extra_cmds(self, value):
        # check if extra_cmds was actually set and return if it wasn't
        if value is None:
            return

        commands = []
        for statement in value:
            for command in statement.cmds:
                commands.append(command)
            commands[-1].end_of_statement = True
        self.commands = commands

    def _get_statements(self):
        statements = []
        current_statement = []

        # set the last command to be the end of the statement
        self.commands[-1].end_of_statement = True
        for command in self.commands:
            # need to use deepcopy as the values can be appended to multiple
            # parents and in lxml that removes it from the original parent,
            # whereas this will create a copy of the statement for each parent
            current_statement.append(deepcopy(command))
            if command.end_of_statement:
                statements.append(current_statement)
                current_statement = []

        return statements


class Command(ComplexObject):
    def __init__(
        self,
        cmd: typing.Optional[str] = None,
        protocol_version: str = "2.3",
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.12 Command
        https://msdn.microsoft.com/en-us/library/dd339976.aspx

        :param cmd: The cmdlet or script to run
        :param protocol_version: The negotiated protocol version of the remote
            host. This determines what merge_* objects are added to the
            serialized xml.
        :param is_script: Whether cmd is a script or not
        :param use_local_scope: Use local or global scope to invoke commands
        :param merge_my_result: Controls the behaviour of what stream to merge
            to 'merge_to_result'. Only supports NONE or ERROR (only used in
            protocol 2.1)
        :param merge_to_result: Controls the behaviour of where to merge the
            'merge_my_result' stream. Only supports NONE or OUTPUT (only used
            in protocol 2.1)
        :param merge_previous: Controls the behaviour of where to merge the
            previous Output and Error streams that have been unclaimed
        :param merge_error: The merge behaviour of the Error stream
        :param merge_warning: The merge behaviour of the Warning stream
        :param merge_verbose: The merge behaviour of the Verbose stream
        :param merge_debug: The merge behaviour of the Debug stream
        :param merge_information: The merge behaviour of the Information stream
        :param args: List of CommandParameters for the cmdlet being invoked
        :param end_of_statement: Whether this command is the last in the
            current statement
        """
        super(Command, self).__init__()
        arg_types = [
            "System.Collections.Generic.List`1[["
            "System.Management.Automation.PSObject, "
            "System.Management.Automation, "
            "Version=1.0.0.0, Culture=neutral, "
            "PublicKeyToken=31bf3856ad364e35]]",
            "System.Object",
        ]
        extended_properties = [
            ("cmd", ObjectMeta("S", name="Cmd")),
            ("is_script", ObjectMeta("B", name="IsScript")),
            ("use_local_scope", ObjectMeta("B", name="UseLocalScope")),
            ("merge_my_result", ObjectMeta("Obj", name="MergeMyResult", object=PipelineResultTypes)),
            ("merge_to_result", ObjectMeta("Obj", name="MergeToResult", object=PipelineResultTypes)),
            ("merge_previous", ObjectMeta("Obj", name="MergePreviousResults", object=PipelineResultTypes)),
            ("args", ListMeta(name="Args", list_value_meta=ObjectMeta(object=CommandParameter), list_types=arg_types)),
        ]

        if version_equal_or_newer(protocol_version, "2.2"):
            extended_properties.extend(
                [
                    ("merge_error", ObjectMeta("Obj", name="MergeError", object=PipelineResultTypes, optional=True)),
                    (
                        "merge_warning",
                        ObjectMeta("Obj", name="MergeWarning", object=PipelineResultTypes, optional=True),
                    ),
                    (
                        "merge_verbose",
                        ObjectMeta("Obj", name="MergeVerbose", object=PipelineResultTypes, optional=True),
                    ),
                    ("merge_debug", ObjectMeta("Obj", name="MergeDebug", object=PipelineResultTypes, optional=True)),
                ]
            )

        if version_equal_or_newer(protocol_version, "2.3"):
            extended_properties.extend(
                [
                    (
                        "merge_information",
                        ObjectMeta("Obj", name="MergeInformation", object=PipelineResultTypes, optional=True),
                    ),
                ]
            )
        self._extended_properties = tuple(extended_properties)

        self.cmd = cmd
        self.protocol_version = protocol_version
        self.is_script = kwargs.get("is_script")
        self.use_local_scope = kwargs.get("use_local_scope")

        none_merge = PipelineResultTypes(value=PipelineResultTypes.NONE)

        # valid in all protocols, only really used in 2.1 (PowerShell 2.0)
        self.merge_my_result = kwargs.get("merge_my_result", none_merge)
        self.merge_to_result = kwargs.get("merge_to_result", none_merge)

        self.merge_previous = kwargs.get("merge_previous", none_merge)

        # only valid for 2.2+ (PowerShell 3.0+)
        self.merge_error = kwargs.get("merge_error", none_merge)
        self.merge_warning = kwargs.get("merge_warning", none_merge)
        self.merge_verbose = kwargs.get("merge_verbose", none_merge)
        self.merge_debug = kwargs.get("merge_debug", none_merge)

        # only valid for 2.3+ (PowerShell 5.0+)
        self.merge_information = kwargs.get("merge_information", none_merge)

        self.args = kwargs.get("args", [])

        # not used in the serialized message but controls how Pipeline is
        # packed (Cmds/ExtraCmds)
        self.end_of_statement = kwargs.get("end_of_statement", False)


class CommandParameter(ComplexObject):
    def __init__(
        self,
        name: typing.Optional[str] = None,
        value: typing.Any = None,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.13 Command Parameter
        https://msdn.microsoft.com/en-us/library/dd359709.aspx

        :param name: The name of the parameter, otherwise None
        :param value: The value of the parameter, can be any primitive type
            or Complex Object, Null for no value
        """
        super(CommandParameter, self).__init__()
        self._extended_properties = (
            ("name", ObjectMeta("S", name="N")),
            ("value", ObjectMeta(name="V")),
        )
        self.name = name
        self.value = value


# The host default data is serialized quite differently from the normal rules
# this contains some sub classes that are specific to the serialized form
class _HostDefaultData(ComplexObject):
    class _DictValue(ComplexObject):
        def __init__(self, **kwargs):
            super(_HostDefaultData._DictValue, self).__init__()
            self._extended_properties = (
                ("value_type", ObjectMeta("S", name="T")),
                ("value", ObjectMeta(name="V")),
            )
            self.value_type = kwargs.get("value_type")
            self.value = kwargs.get("value")

    class _Color(ComplexObject):
        def __init__(self, color):
            super(_HostDefaultData._Color, self).__init__()
            self._extended_properties = (
                ("type", ObjectMeta("S", name="T")),
                ("color", ObjectMeta("I32", name="V")),
            )
            self.type = "System.ConsoleColor"
            self.color = color.value

    class _Coordinates(ComplexObject):
        def __init__(self, coordinates):
            super(_HostDefaultData._Coordinates, self).__init__()
            self._extended_properties = (
                ("type", ObjectMeta("S", name="T")),
                ("value", ObjectMeta("ObjDynamic", name="V", object=GenericComplexObject)),
            )
            self.type = "System.Management.Automation.Host.Coordinates"
            self.value = GenericComplexObject()
            self.value.extended_properties["x"] = coordinates.x
            self.value.extended_properties["y"] = coordinates.y

    class _Size(ComplexObject):
        def __init__(self, size):
            super(_HostDefaultData._Size, self).__init__()
            self._extended_properties = (
                ("type", ObjectMeta("S", name="T")),
                ("value", ObjectMeta("ObjDynamic", name="V", object=GenericComplexObject)),
            )
            self.type = "System.Management.Automation.Host.Size"
            self.value = GenericComplexObject()
            self.value.extended_properties["width"] = size.width
            self.value.extended_properties["height"] = size.height

    def __init__(self, **kwargs):
        # Used by HostInfo to encapsulate the host info values inside a
        # special object required by PSRP
        super(_HostDefaultData, self).__init__()
        key_meta = ObjectMeta("I32", name="Key")
        self._extended_properties = (("_host_dict", DictionaryMeta(name="data", dict_key_meta=key_meta)),)
        self.raw_ui = kwargs.get("raw_ui")

    @property
    def _host_dict(self):
        return (
            (0, self._Color(self.raw_ui.foreground_color)),
            (1, self._Color(self.raw_ui.background_color)),
            (2, self._Coordinates(self.raw_ui.cursor_position)),
            (3, self._Coordinates(self.raw_ui.window_position)),
            (4, self._DictValue(value_type="System.Int32", value=self.raw_ui.cursor_size)),
            (5, self._Size(self.raw_ui.buffer_size)),
            (6, self._Size(self.raw_ui.window_size)),
            (7, self._Size(self.raw_ui.max_window_size)),
            (8, self._Size(self.raw_ui.max_physical_window_size)),
            (9, self._DictValue(value_type="System.String", value=self.raw_ui.window_title)),
        )


class HostInfo(ComplexObject):
    def __init__(
        self,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.14 HostInfo
        https://msdn.microsoft.com/en-us/library/dd340936.aspx

        :param host: An implementation of pypsrp.host.PSHost that defines the
            local host
        """
        super(HostInfo, self).__init__()
        self._extended_properties = (
            ("_host_data", ObjectMeta("Obj", name="_hostDefaultData", optional=True, object=_HostDefaultData)),
            ("_is_host_null", ObjectMeta("B", name="_isHostNull")),
            ("_is_host_ui_null", ObjectMeta("B", name="_isHostUINull")),
            ("_is_host_raw_ui_null", ObjectMeta("B", name="_isHostRawUINull")),
            ("_use_runspace_host", ObjectMeta("B", name="_useRunspaceHost")),
        )
        self.host = kwargs.get("host", None)

    @property
    def _is_host_null(self):
        return self.host is None

    @property
    def _is_host_ui_null(self):
        if self.host is not None:
            return self.host.ui is None
        else:
            return True

    @property
    def _is_host_raw_ui_null(self):
        if self.host is not None and self.host.ui is not None:
            return self.host.ui.raw_ui is None
        else:
            return True

    @property
    def _use_runspace_host(self):
        return self.host is None

    @property
    def _host_data(self):
        if self._is_host_raw_ui_null:
            return None
        else:
            host_data = _HostDefaultData(raw_ui=self.host.ui.raw_ui)
            return host_data


class ErrorRecord(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.15 ErrorRecord
        https://msdn.microsoft.com/en-us/library/dd340106.aspx
        """
        super(ErrorRecord, self).__init__()
        self._types = ["System.Management.Automation.ErrorRecord", "System.Object"]
        self._extended_properties = (
            ("exception", ObjectMeta(name="Exception", optional=True)),
            ("target_object", ObjectMeta(name="TargetObject", optional=True)),
            ("invocation", ObjectMeta("B", name="SerializeExtendedInfo")),
            (
                "invocation_info",
                ObjectMeta("ObjDynamic", name="InvocationInfo", object=GenericComplexObject, optional=True),
            ),
            ("fq_error", ObjectMeta("S", name="FullyQualifiedErrorId")),
            ("category", ObjectMeta("I32", name="ErrorCategory_Category")),
            ("activity", ObjectMeta("S", name="ErrorCategory_Activity", optional=True)),
            ("reason", ObjectMeta("S", name="ErrorCategory_Reason", optional=True)),
            ("target_name", ObjectMeta("S", name="ErrorCategory_TargetName", optional=True)),
            ("target_type", ObjectMeta("S", name="ErrorCategory_TargetType", optional=True)),
            ("message", ObjectMeta("S", name="ErrorCategory_Message", optional=True)),
            ("details_message", ObjectMeta("S", name="ErrorDetails_Message", optional=True)),
            ("action", ObjectMeta("S", name="ErrorDetails_RecommendedAction", optional=True)),
            ("script_stacktrace", ObjectMeta("S", name="ErrorDetails_ScriptStackTrace", optional=True)),
            ("extended_info_present", ObjectMeta("B", name="SerializeExtendedInfo")),
            ("invocation_name", ObjectMeta("S", optional=True, name="InvocationInfo_InvocationName")),
            (
                "invocation_bound_parameters",
                DictionaryMeta(
                    name="InvocationInfo_BoundParameters",
                    optional=True,
                    dict_key_meta=ObjectMeta("S"),
                    dict_types=[
                        "System.Management.Automation.PSBoundParametersDictionary",
                        "System.Collections.Generic.Dictionary`2[[System.String, "
                        "mscorlib, Version=4.0.0.0, Culture=neutral, "
                        "PublicKeyToken=b77a5c561934e089],"
                        "[System.Object, mscorlib, Version=4.0.0.0, "
                        "Culture=neutral, PublicKeyToken=b77a5c561934e089]]",
                        "System.Object",
                    ],
                ),
            ),
            (
                "invocation_unbound_arguments",
                ListMeta(
                    name="InvocationInfo_UnboundArguments",
                    optional=True,
                    list_types=[
                        "System.Collections.Generic.List`1[["
                        "System.Object, mscorlib, Version=4.0.0.0, "
                        "Culture=neutral, PublicKeyToken=b77a5c561934e089]]",
                        "System.Object",
                    ],
                ),
            ),
            (
                "invocation_command_origin",
                ObjectMeta("Obj", name="InvocationInfo_CommandOrigin", optional=True, object=CommandOrigin),
            ),
            ("invocation_expecting_input", ObjectMeta("B", name="InvocationInfo_ExpectingInput", optional=True)),
            ("invocation_line", ObjectMeta("S", name="InvocationInfo_Line", optional=True)),
            ("invocation_offset_in_line", ObjectMeta("I32", name="InvocationInfo_OffsetInLine", optional=True)),
            ("invocation_position_message", ObjectMeta("S", name="InvocationInfo_PositionMessage", optional=True)),
            ("invocation_script_name", ObjectMeta("S", name="InvocationInfo_ScriptName", optional=True)),
            ("invocation_script_line_number", ObjectMeta("I32", name="InvocationInfo_ScriptLineNumber", optional=True)),
            ("invocation_history_id", ObjectMeta("I64", name="InvocationInfo_HistoryId", optional=True)),
            ("invocation_pipeline_length", ObjectMeta("I32", name="InvocationInfo_PipelineLength", optional=True)),
            ("invocation_pipeline_position", ObjectMeta("I32", name="InvocationInfo_PipelinePosition", optional=True)),
            (
                "invocation_pipeline_iteration_info",
                ListMeta(
                    name="InvocationInfo_PipelineIterationInfo",
                    optional=True,
                    list_value_meta=ObjectMeta("I32"),
                    list_types=["System.In32[]", "System.Array", "System.Object"],
                ),
            ),
            (
                "command_type",
                ObjectMeta(
                    "Obj",
                    name="CommandInfo_CommandType",
                    object=CommandType,
                    optional=True,
                ),
            ),
            (
                "command_definition",
                ObjectMeta(
                    "S",
                    name="CommandInfo_Definition",
                    optional=True,
                ),
            ),
            ("command_name", ObjectMeta("S", name="CommandInfo_Name", optional=True)),
            (
                "command_visibility",
                ObjectMeta("Obj", name="CommandInfo_Visibility", object=SessionStateEntryVisibility, optional=True),
            ),
            (
                "pipeline_iteration_info",
                ListMeta(
                    name="PipelineIterationInfo",
                    optional=True,
                    list_value_meta=ObjectMeta("I32"),
                    list_types=[
                        "System.Collections.ObjectModel.ReadOnlyCollection`1[["
                        "System.Int32, mscorlib, Version=4.0.0.0, "
                        "Culture=neutral, PublicKeyToken=b77a5c561934e089]]",
                        "System.Object",
                    ],
                ),
            ),
        )
        self.exception = kwargs.get("exception")
        self.target_info = kwargs.get("target_info")
        self.invocation = kwargs.get("invocation")
        self.fq_error = kwargs.get("fq_error")
        self.category = kwargs.get("category")
        self.activity = kwargs.get("activity")
        self.reason = kwargs.get("reason")
        self.target_name = kwargs.get("target_name")
        self.target_type = kwargs.get("target_type")
        self.message = kwargs.get("message")
        self.details_message = kwargs.get("details_message")
        self.action = kwargs.get("action")
        self.pipeline_iteration_info = kwargs.get("pipeline_iteration_info")
        self.invocation_name = kwargs.get("invocation_name")
        self.invocation_bound_parameters = kwargs.get("invocation_bound_parameters")
        self.invocation_unbound_arguments = kwargs.get("invocation_unbound_arguments")
        self.invocation_command_origin = kwargs.get("invocation_command_origin")
        self.invocation_expecting_input = kwargs.get("invocation_expecting_input")
        self.invocation_line = kwargs.get("invocation_line")
        self.invocation_offset_in_line = kwargs.get("invocation_offset_in_line")
        self.invocation_position_message = kwargs.get("invocation_position_message")
        self.invocation_script_name = kwargs.get("invocation_script_name")
        self.invocation_script_line_number = kwargs.get("invocation_script_line_number")
        self.invocation_history_id = kwargs.get("invocation_history_id")
        self.invocation_pipeline_length = kwargs.get("invocation_pipeline_length")
        self.invocation_pipeline_position = kwargs.get("invocation_pipeline_position")
        self.invocation_pipeline_iteration_info = kwargs.get("invocation_pipeline_iteration_info")
        self.command_type = kwargs.get("command_type")
        self.command_definition = kwargs.get("command_definition")
        self.command_name = kwargs.get("command_name")
        self.command_visibility = kwargs.get("command_visibility")
        self.extended_info_present = self.invocation is not None


class InformationalRecord(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.16 InformationalRecord (Debug/Warning/Verbose)
        https://msdn.microsoft.com/en-us/library/dd305072.aspx
        """
        super(InformationalRecord, self).__init__()
        self._types = ["System.Management.Automation.InformationRecord", "System.Object"]
        self._extended_properties = (
            ("message", ObjectMeta("S", name="InformationalRecord_Message")),
            ("invocation", ObjectMeta("B", name="InformationalRecord_SerializeInvocationInfo")),
            ("invocation_name", ObjectMeta("S", optional=True, name="InvocationInfo_InvocationName")),
            (
                "invocation_bound_parameters",
                DictionaryMeta(
                    name="InvocationInfo_BoundParameters",
                    optional=True,
                    dict_key_meta=ObjectMeta("S"),
                    dict_types=[
                        "System.Management.Automation.PSBoundParametersDictionary",
                        "System.Collections.Generic.Dictionary`2[[System.String, "
                        "mscorlib, Version=4.0.0.0, Culture=neutral, "
                        "PublicKeyToken=b77a5c561934e089],"
                        "[System.Object, mscorlib, Version=4.0.0.0, "
                        "Culture=neutral, PublicKeyToken=b77a5c561934e089]]",
                        "System.Object",
                    ],
                ),
            ),
            (
                "invocation_unbound_arguments",
                ListMeta(
                    name="InvocationInfo_UnboundArguments",
                    optional=True,
                    list_types=[
                        "System.Collections.Generic.List`1[["
                        "System.Object, mscorlib, Version=4.0.0.0, "
                        "Culture=neutral, PublicKeyToken=b77a5c561934e089]]",
                        "System.Object",
                    ],
                ),
            ),
            (
                "invocation_command_origin",
                ObjectMeta("Obj", name="InvocationInfo_CommandOrigin", optional=True, object=CommandOrigin),
            ),
            ("invocation_expecting_input", ObjectMeta("B", name="InvocationInfo_ExpectingInput", optional=True)),
            ("invocation_line", ObjectMeta("S", name="InvocationInfo_Line", optional=True)),
            ("invocation_offset_in_line", ObjectMeta("I32", name="InvocationInfo_OffsetInLine", optional=True)),
            ("invocation_position_message", ObjectMeta("S", name="InvocationInfo_PositionMessage", optional=True)),
            ("invocation_script_name", ObjectMeta("S", name="InvocationInfo_ScriptName", optional=True)),
            ("invocation_script_line_number", ObjectMeta("I32", name="InvocationInfo_ScriptLineNumber", optional=True)),
            ("invocation_history_id", ObjectMeta("I64", name="InvocationInfo_HistoryId", optional=True)),
            ("invocation_pipeline_length", ObjectMeta("I32", name="InvocationInfo_PipelineLength", optional=True)),
            ("invocation_pipeline_position", ObjectMeta("I32", name="InvocationInfo_PipelinePosition", optional=True)),
            (
                "invocation_pipeline_iteration_info",
                ListMeta(
                    name="InvocationInfo_PipelineIterationInfo",
                    optional=True,
                    list_value_meta=ObjectMeta("I32"),
                    list_types=["System.In32[]", "System.Array", "System.Object"],
                ),
            ),
            (
                "command_type",
                ObjectMeta(
                    "Obj",
                    name="CommandInfo_CommandType",
                    object=CommandType,
                    optional=True,
                ),
            ),
            (
                "command_definition",
                ObjectMeta(
                    "S",
                    name="CommandInfo_Definition",
                    optional=True,
                ),
            ),
            ("command_name", ObjectMeta("S", name="CommandInfo_Name", optional=True)),
            (
                "command_visibility",
                ObjectMeta("Obj", name="CommandInfo_Visibility", object=SessionStateEntryVisibility, optional=True),
            ),
            (
                "pipeline_iteration_info",
                ListMeta(
                    name="InformationalRecord_PipelineIterationInfo",
                    optional=True,
                    list_value_meta=ObjectMeta("I32"),
                    list_types=[
                        "System.Collections.ObjectModel.ReadOnlyCollection`1[["
                        "System.Int32, mscorlib, Version=4.0.0.0, "
                        "Culture=neutral, PublicKeyToken=b77a5c561934e089]]",
                        "System.Object",
                    ],
                ),
            ),
        )
        self.message = kwargs.get("message")
        self.pipeline_iteration_info = kwargs.get("pipeline_iteration_info")
        self.invocation_name = kwargs.get("invocation_name")
        self.invocation_bound_parameters = kwargs.get("invocation_bound_parameters")
        self.invocation_unbound_arguments = kwargs.get("invocation_unbound_arguments")
        self.invocation_command_origin = kwargs.get("invocation_command_origin")
        self.invocation_expecting_input = kwargs.get("invocation_expecting_input")
        self.invocation_line = kwargs.get("invocation_line")
        self.invocation_offset_in_line = kwargs.get("invocation_offset_in_line")
        self.invocation_position_message = kwargs.get("invocation_position_message")
        self.invocation_script_name = kwargs.get("invocation_script_name")
        self.invocation_script_line_number = kwargs.get("invocation_script_line_number")
        self.invocation_history_id = kwargs.get("invocation_history_id")
        self.invocation_pipeline_length = kwargs.get("invocation_pipeline_length")
        self.invocation_pipeline_position = kwargs.get("invocation_pipeline_position")
        self.invocation_pipeline_iteration_info = kwargs.get("invocation_pipeline_iteration_info")
        self.command_type = kwargs.get("command_type")
        self.command_definition = kwargs.get("command_definition")
        self.command_name = kwargs.get("command_name")
        self.command_visibility = kwargs.get("command_visibility")
        self.invocation = False


class HostMethodIdentifier(Enum):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.17 Host Method Identifier
        https://msdn.microsoft.com/en-us/library/dd306624.aspx

        Represents methods to be executed on a host.

        :param value: The method identifier to execute
        """
        string_map = {
            1: "GetName",
            2: "GetVersion",
            3: "GetInstanceId",
            4: "GetCurrentCulture",
            5: "GetCurrentUICulture",
            6: "SetShouldExit",
            7: "EnterNestedPrompt",
            8: "ExitNestedPrompt",
            9: "NotifyBeginApplication",
            10: "NotifyEndApplication",
            11: "ReadLine",
            12: "ReadLineAsSecureString",
            13: "Write1",
            14: "Write2",
            15: "WriteLine1",
            16: "WriteLine2",
            17: "WriteLine3",
            18: "WriteErrorLine",
            19: "WriteDebugLine",
            20: "WriteProgress",
            21: "WriteVerboseLine",
            22: "WriteWarningLine",
            23: "Prompt",
            24: "PromptForCredential1",
            25: "PromptForCredential2",
            26: "PromptForChoice",
            27: "GetForegroundColor",
            28: "SetForegroundColor",
            29: "GetBackgroundColor",
            30: "SetBackgroundColor",
            31: "GetCursorPosition",
            32: "SetCursorPosition",
            33: "GetWindowPosition",
            34: "SetWindowPosition",
            35: "GetCursorSize",
            36: "SetCursorSize",
            37: "GetBufferSize",
            38: "SetBufferSize",
            39: "GetWindowSize",
            40: "SetWindowSize",
            41: "GetWindowTitle",
            42: "SetWindowTitle",
            43: "GetMaxWindowSize",
            44: "GetMaxPhysicalWindowSize",
            45: "GetKeyAvailable",
            46: "ReadKey",
            47: "FlushInputBuffer",
            48: "SetBufferContents1",
            49: "SetBufferContents2",
            50: "GetBufferContents",
            51: "ScrollBufferContents",
            52: "PushRunspace",
            53: "PopRunspace",
            54: "GetIsRunspacePushed",
            55: "GetRunspce",
            56: "PromptForChoiceMultipleSelection",
        }
        super(HostMethodIdentifier, self).__init__(
            "System.Management.Automation.Remoting.RemoteHostMethodId", string_map, **kwargs
        )


class CommandType(Enum):
    ALIAS = 0x0001
    FUNCTION = 0x0002
    FILTER = 0x0004
    CMDLET = 0x0008
    EXTERNAL_SCRIPT = 0x0010
    APPLICATION = 0x0020
    SCRIPT = 0x0040
    WORKFLOW = 0x0080
    CONFIGURATION = 0x0100
    ALL = 0x01FF

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.19 CommandType
        https://msdn.microsoft.com/en-us/library/ee175965.aspx

        :param value: The initial flag value for CommandType
        """
        super(CommandType, self).__init__("System.Management.Automation.CommandTypes", {}, **kwargs)

    @property  # type: ignore[override]
    def _to_string(self) -> str:
        if self.value == 0x01FF:
            return "All"

        string_map = (
            ("Alias", 0x0001),
            ("Function", 0x0002),
            ("Filter", 0x0004),
            ("Cmdlet", 0x0008),
            ("ExternalScript", 0x0010),
            ("Application", 0x0020),
            ("Script", 0x0040),
            ("Workflow", 0x0080),
            ("Configuration", 0x0100),
        )
        values = []
        for name, flag in string_map:
            if (self.value or 0) & flag == flag:
                values.append(name)
        return ", ".join(values)

    @_to_string.setter
    def _to_string(self, value):
        pass


class CommandMetadataCount(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.21 CommandMetadataCount
        https://msdn.microsoft.com/en-us/library/ee175881.aspx

        :param count: The number of CommandMetadata messages in the pipeline
            output
        """
        super(CommandMetadataCount, self).__init__()
        self.types = [
            "Selected.Microsoft.PowerShell.Commands.GenericMeasureInfo",
            "System.Management.Automation.PSCustomObject",
            "System.Object",
        ]
        self._extended_properties = (("count", ObjectMeta("I32", name="Count")),)
        self.count = kwargs.get("count")


class CommandMetadata(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.22 CommandMetadata
        https://msdn.microsoft.com/en-us/library/ee175993.aspx

        :param name: The name of a command
        :param namespace: The namespace of the command
        :param help_uri: The URI to the documentation of the command
        :param command_type: The CommandType of the command
        :param output_type: The types of objects that a command can send as
            output
        :param parameters: Metadata of parameters that the command can accept
            as Command Parameters
        """
        super(CommandMetadata, self).__init__()
        self.types = ["System.Management.Automation.PSCustomObject", "System.Object"]
        self._extended_properties = (
            ("name", ObjectMeta("S", name="Name")),
            ("namespace", ObjectMeta("S", name="Namespace")),
            ("help_uri", ObjectMeta("S", name="HelpUri")),
            ("command_type", ObjectMeta("Obj", name="CommandType", object=CommandType)),
            (
                "output_type",
                ListMeta(
                    name="OutputType",
                    list_value_meta=ObjectMeta("S"),
                    list_types=[
                        "System.Collections.ObjectModel.ReadOnlyCollection`1[["
                        "System.Management.Automation.PSTypeName, "
                        "System.Management.Automation, Version=3.0.0.0, "
                        "Culture=neutral, PublicKeyToken=31bf3856ad364e35]]",
                    ],
                ),
            ),
            (
                "parameters",
                DictionaryMeta(
                    name="Parameters",
                    dict_key_meta=ObjectMeta("S"),
                    dict_value_meta=ObjectMeta("Obj", object=ParameterMetadata),
                ),
            ),
        )
        self.name = kwargs.get("name")
        self.namespace = kwargs.get("namespace")
        self.help_uri = kwargs.get("help_uri")
        self.command_type = kwargs.get("command_type")
        self.output_type = kwargs.get("output_type")
        self.parameters = kwargs.get("parameters")


class ParameterMetadata(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.23 ParameterMetadata
        https://msdn.microsoft.com/en-us/library/ee175918.aspx

        :param name: The name of a parameter
        :param parameter_type: The type of the parameter
        :param alises: List of alternative names of the parameter
        :param switch_parameter: True if param is a switch parameter
        :param dynamic: True if param is included as a consequence of the data
            specified in the ArgumentList property
        """
        super(ParameterMetadata, self).__init__()
        self.types = ["System.Management.Automation.ParameterMetadata", "System.Object"]
        self._adapted_properties = (
            ("name", ObjectMeta("S", name="Name")),
            ("parameter_type", ObjectMeta("S", name="ParameterType")),
            (
                "aliases",
                ListMeta(
                    name="Aliases",
                    list_value_meta=ObjectMeta("S"),
                    list_types=[
                        "System.Collections.ObjectModel.Collection`1"
                        "[[System.String, mscorlib, Version=4.0.0.0, "
                        "Culture=neutral, PublicKeyToken=b77a5c561934e089]]",
                        "System.Object",
                    ],
                ),
            ),
            ("switch_parameter", ObjectMeta("B", name="SwitchParameter")),
            ("dynamic", ObjectMeta("B", name="IsDynamic")),
        )
        self.name = kwargs.get("name")
        self.parameter_type = kwargs.get("parameter_type")
        self.aliases = kwargs.get("aliases")
        self.switch_parameter = kwargs.get("switch_parameter")
        self.dynamic = kwargs.get("dynamic")


class PSCredential(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.25 PSCredential
        https://msdn.microsoft.com/en-us/library/ee442231.aspx

        Represents a username and a password. As the password is a secure
        string, the RunspacePool must have already exchanged keys with
        .exchange_keys() method.

        :param username: The username (including the domain if required)
        :param password: The password for the user, this should be a unicode
            string in order to make sure the encoding is correct
        """
        super(PSCredential, self).__init__()
        self._types = ["System.Management.Automation.PSCredential", "System.Object"]
        self._adapted_properties = (
            ("username", ObjectMeta("S", name="UserName")),
            ("password", ObjectMeta("SS", name="Password")),
        )
        self._to_string = "System.Management.Automation.PSCredential"

        self.username = kwargs.get("username")
        self.password = kwargs.get("password")


class KeyInfo(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.26 KeyInfo
        https://msdn.microsoft.com/en-us/library/ee441795.aspx

        Represents information about a keyboard event, this is used for the
        serialized of a ReadKey host method and is not the same as the
        serialized form of KeyInfo in .NET (see KeyInfoDotNet).

        :param code: The int value for the virtual key code
        :param character: The character
        :param state: The ControlKeyState int value
        :param key_down: Whether the key is pressed or released
        """
        super(KeyInfo, self).__init__()
        self._extended_properties = (
            ("code", ObjectMeta("I32", name="virtualKeyCode", optional=True)),
            ("character", ObjectMeta("C", name="character")),
            ("state", ObjectMeta("I32", name="controlKeyState")),
            ("key_down", ObjectMeta("B", name="keyDown")),
        )
        self.code = kwargs.get("code")
        self.character = kwargs.get("character")
        self.state = kwargs.get("state")
        self.key_down = kwargs.get("key_down")


class KeyInfoDotNet(ComplexObject):
    def __init__(self, **kwargs):
        """
        System.Management.Automation.Host.KeyInfo
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.keyinfo

        This is the proper serialized form of KeyInfo from .NET, it is
        returned in a PipelineOutput message.

        :param code: The int value for the virtual key code
        :param character: The character
        :param state: The ControlKeyState as a string value
        :param key_down: Whether the key is pressed or released
        """
        super(KeyInfoDotNet, self).__init__()
        self._types = ["System.Management.Automation.Host.KeyInfo", "System.ValueType", "System.Object"]
        self._adapted_properties = (
            ("code", ObjectMeta("I32", name="VirtualKeyCode")),
            ("character", ObjectMeta("C", name="Character")),
            ("state", ObjectMeta("S", name="ControlKeyState")),
            ("key_down", ObjectMeta("B", name="KeyDown")),
        )
        self.code = kwargs.get("code")
        self.character = kwargs.get("character")
        self.state = kwargs.get("state")
        self.key_down = kwargs.get("key_down")


class ControlKeyState(object):
    """
    [MS-PSRP] 2.2.3.27 ControlKeyStates
    https://msdn.microsoft.com/en-us/library/ee442685.aspx
    https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.controlkeystates

    A set of zero or more control keys that are help down.
    """

    RightAltPressed = 0x0001
    LeftAltPressed = 0x0002
    RightCtrlPressed = 0x0004
    LeftCtrlPressed = 0x0008
    ShiftPressed = 0x0010
    NumLockOn = 0x0020
    ScrollLockOn = 0x0040
    CapsLockOn = 0x0080
    EnhancedKey = 0x0100


class BufferCell(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.3.28 BufferCell
        https://msdn.microsoft.com/en-us/library/ee443291.aspx

        The contents of a cell of a host's screen buffer.

        :param character: The chracter visibile in the cell
        :param foreground_color: The Color of the foreground
        :param background_color: The Color of the background
        :param cell_type: The int value of BufferCellType
        """
        super(BufferCell, self).__init__()
        self._adapted_properties = (
            ("character", ObjectMeta("C", name="character")),
            ("foreground_color", ObjectMeta("Obj", name="foregroundColor", object=Color)),
            ("background_color", ObjectMeta("Obj", name="backgroundColor", object=Color)),
            ("cell_type", ObjectMeta("I32", name="bufferCellType")),
        )
        self.character = kwargs.get("character")
        self.foreground_color = kwargs.get("foreground_color")
        self.background_color = kwargs.get("background_color")
        self.cell_type = kwargs.get("cell_type")


class BufferCellType(object):
    """
    [MS-PSRP] 2.2.3.29 BufferCellType
    https://msdn.microsoft.com/en-us/library/ee442184.aspx

    The type of a cell of a screen buffer.
    """

    COMPLETE = 0
    LEADING = 1
    TRAILING = 2


class Array(ComplexObject):
    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.6.1.4 Array
        https://msdn.microsoft.com/en-us/library/dd340684.aspx

        Represents a (potentially multi-dimensional) array of elements.

        :param array: The array (list) that needs to be serialised. This can
            be a multidimensional array (lists in a list)
        """
        super(Array, self).__init__()
        self._extended_properties = (
            ("mae", ListMeta(name="mae")),
            ("mal", ListMeta(name="mal", list_value_meta=ObjectMeta("I32"))),
        )
        self._array = None
        self._mae = None
        self._mal = None
        self._array = kwargs.get("array")

    @property
    def array(self):
        if self._array is None:
            self._array = self._build_array(self._mae, self._mal)

        return self._array

    @array.setter
    def array(self, value):
        self._array = value

    @property
    def mae(self):
        # elements of the array are flattened into a list and ordered by first
        # listing the deepest elements
        mae = self._get_list_entries(self._array)
        return mae

    @mae.setter
    def mae(self, value):
        self._mae = value

    @property
    def mal(self):
        mal = self._get_list_count(self.array)
        return mal

    @mal.setter
    def mal(self, value):
        self._mal = value

    def _build_array(self, mae, mal):
        values = []

        length = mal.pop(-1)
        while True:
            entry = []
            for i in range(0, length):
                entry.append(mae.pop(0))
            values.append(entry)
            if len(mae) == 0:
                break

        if len(mal) == 0:
            values = values[0]
        elif len(mal) > 1:
            values = self._build_array(values, mal)

        return values

    def _get_list_entries(self, list_value):
        values = []
        for value in list_value:
            if isinstance(value, list):
                values.extend(self._get_list_entries(value))
            else:
                values.append(value)

        return values

    def _get_list_count(self, list_value):
        count = []

        current_entry = list_value
        while True:
            if isinstance(current_entry, list):
                count.append(len(current_entry))
                current_entry = current_entry[0]
            else:
                break

        return count


class CommandOrigin(Enum):
    RUNSPACE = 0
    INTERNAL = 1

    def __init__(self, **kwargs):
        """
        [MS-PSRP] 2.2.2.30 CommandOrigin
        https://msdn.microsoft.com/en-us/library/ee441964.aspx

        :param value: The command origin flag to set
        """
        string_map = {
            0: "Runspace",
            1: "Internal",
        }
        super(CommandOrigin, self).__init__("System.Management.Automation.CommandOrigin", string_map, **kwargs)


class PipelineResultTypes(Enum):
    # While MS-PSRP show this as flags with different values, we only
    # ever send NONE OUTPUT, ERROR, or OUTPUT_AND_ERROR across the wire and
    # the actual C# code use these values as enums and not flags. We will
    # replicate that behaviour here. If using these values, do not rely on
    # the actual numeric values but rather these definitions.

    NONE = 0  # default streaming behaviour
    OUTPUT = 1
    ERROR = 2
    WARNING = 3  # also output and error for MergePreviousResults (PS v2)
    VERBOSE = 4
    DEBUG = 5
    INFORMATION = 6
    ALL = 7  # Error, Warning, Verbose, Debug, Information streams
    NULL = 8  # redirect to nothing - pretty much the same as null

    def __init__(
        self,
        protocol_version_2: bool = False,
        **kwargs: typing.Any,
    ) -> None:
        """
        [MS-PSRP] 2.2.3.31 PipelineResultTypes
        https://msdn.microsoft.com/en-us/library/ee938207.aspx

        Used as identifiers

        :param protocol_version_2: Whether to use the original string map or
            just None, Output, and Error that are a bitwise combination. This
            is only really relevant for MergePreviousResults in a Command obj
        :param value: The initial PipelineResultType flag to set
        """
        if protocol_version_2 is True:
            string_map = {
                0: "None",
                1: "Output",
                2: "Error",
                3: "Output, Error",
            }
        else:
            string_map = {
                0: "None",
                1: "Output",
                2: "Error",
                3: "Warning",
                4: "Verbose",
                5: "Debug",
                6: "Information",
                7: "All",
                8: "Null",
            }
        super(PipelineResultTypes, self).__init__(
            "System.Management.Automation.Runspaces.PipelineResultTypes", string_map, **kwargs
        )


class CultureInfo(ComplexObject):
    def __init__(self, **kwargs):
        super(CultureInfo, self).__init__()

        self._adapted_properties = (
            ("lcid", ObjectMeta("I32", name="LCID")),
            ("name", ObjectMeta("S", name="Name")),
            ("display_name", ObjectMeta("S", name="DisplayName")),
            ("ietf_language_tag", ObjectMeta("S", name="IetfLanguageTag")),
            ("three_letter_iso_name", ObjectMeta("S", name="ThreeLetterISOLanguageName")),
            ("three_letter_windows_name", ObjectMeta("S", name="ThreeLetterWindowsLanguageName")),
            ("two_letter_iso_language_name", ObjectMeta("S", name="TwoLetterISOLanguageName")),
        )
        self.lcid = kwargs.get("lcid")
        self.name = kwargs.get("name")
        self.display_name = kwargs.get("display_name")
        self.ieft_language_tag = kwargs.get("ietf_language_tag")
        self.three_letter_iso_name = kwargs.get("three_letter_iso_name")
        self.three_letter_windows_name = kwargs.get("three_letter_windows_name")
        self.two_letter_iso_language_name = kwargs.get("two_letter_iso_language_name")


class ProgressRecordType(Enum):
    PROCESSING = 0
    COMPLETED = 1

    def __init__(self, **kwargs):
        """
        System.Management.Automation.ProgressRecordType Enum
        This isn't in MS-PSRP but is used in the InformationRecord message and
        so we need to define it here.

        :param value: The initial ProgressRecordType value to set
        """
        string_map = {
            0: "Processing",
            1: "Completed",
        }
        super(ProgressRecordType, self).__init__(
            "System.Management.Automation.ProgressRecordType", string_map, **kwargs
        )


class SessionStateEntryVisibility(Enum):
    PUBLIC = 0
    PRIVATE = 1

    def __init__(self, **kwargs):
        """
        System.Management.Automation.SessionStateEntryVisibility Enum
        This isn't in MS-PSRP but is used in the InformationalRecord object so
        we need to define it here

        :param value: The initial SessionStateEntryVisibility value to set
        """
        string_map = {0: "Public", 1: "Private"}
        super(SessionStateEntryVisibility, self).__init__(
            "System.Management.Automation.SessionStateEntryVisibility", string_map, **kwargs
        )
